# HISTORY COMMENTS (Thomas Erben)
#
# 01.10.2006:
# Changes to reflect upgrades to new versions of
# external packages:
# - CFITSIO: 2.4 -> 2.51
# - PGPLOT: 5.2 -> 5.22
#
# Perl libraries:
# - ExtUtils: 1.13 -> 1.16 (unofficial version of ExtUtils)
# - Filter: 1.29 -> 1.34
# - Text-Balanced: 1.95 -> 2.00
# - PDL: 2.40 -> 2.43
# - Tk: 800.023 -> 804.028
# - PGPLOT: 2.18 -> 2.19
# - Astro-FITS-CFITSIO: 1.0 -> 1.05
# - String-Approx: 3.15 -> 3.26
# - TermReadKey: 2.14 -> 2.30
#
# 13.06.2007:
# Upgrade of external packages:
# - cdsclient: 2.86 -> 2.87 (scripts to request SDSS 4 and SDSS 5)
# - CFITSIO: 2.51 -> 3.03   (2.51 does not work with Astro-FITS-CFITSIO
#                            on Windows/CYGWIN)
#
# Update of Perl libraries:
# - PGPLOT: 2.19 -> 2.20    (no special reason for the update! I came
#                            across this package by chance when testing
#                            WIFIX installation on Windows/CYGWIN)
#
# 27.01.2008:
# Upgrade of external packages:
# - cdsclient: 2.87 -> 2.89 
# - CFITSIO: 3.03 -> 3.06   
#                           
# Upgrade of Perl libraries:
# - Tk: 800.025 -> 804.028       (Tk 800.025 does not compile with Perl 5.10)
# - Filter: 1.32 -> 1.34
# - Text-Balanced: 1.98 -> 2.00
#
# 06.05.2008:
# cdsclient: 2.89 -> 3.0
#
# 03.06.2008:
# The default install directory is now $ENV{HOME}/PipeLibs
#
# 09.04.2008:
# An upgrade of cfitsio was necessary to avoid compiler warnings with g++-4.3
# beacuse of type-unsafe conversions. Other Perl modules were brought
# to the newest versions:
#
# Upgrade of Perl libraries:
# - Filter: 1.34 -> 1.36
# - PDL: 2.3 -> 2.4
#
# Upgrade of external libraries:
# - cfitsio: 3.06 -> 3.14
# 
# 06.12.2009:
# Upgrade of external programs:
# - cdsclient: 3.0 -> 3.4
# - cfitsio: 3.14 _> 3.21
# 
# Upgrade of Perl modules:
# - Filter: 1.36 -> 1.37
# - PDL: 2.4.4 -> 2.4.5
# - Text-Balanced: 2.0 -> 2.02

use strict;
use vars qw(@BuildConf %BuildCVals);

my $BuildConf='Build.conf';

sub InitBuild {

 @BuildConf=(
 
    {
   KEY  =>  'INSTALL',
   COMM =>  ' The directory where the stuff will be installed',
    }, 


    {
   COMM =>  ' ==== The external libraries/packages',
    }, 


   {
   KEY  =>  'mkEXTERN',
   COMM =>  ' Install external libraries [y/n] ?',
    }, 


   {
   KEY  =>  'CFITSIO',
   COMM =>  ' The name of the CFITSIO library',
    }, 

   {
   KEY  =>  'CFITSIO_LIB',
   COMM =>  ' If already installed, where is libcfitsio.a',
    }, 

   {
   KEY  =>  'CFITSIO_INC',
   COMM =>  ' If already installed, where are the include files',
    }, 

   {
   KEY  =>  'PGPLOT',
   COMM =>  ' The name of the PGPLOT library',
    }, 

   {
   KEY  =>  'PGPLOT_LIB',
   COMM =>  ' If already installed, where is libpgplot.a',
    }, 


   {
   KEY  =>  'CDSCLIENT',
   COMM =>  ' The name of the CDSCLIENT library',
    }, 


    {
   COMM =>  ' ==== The PERL modules',
    }, 


   {
   KEY  =>  'mkPERL',
   COMM =>  ' Install the PERL modules [y/n] ?',
    }, 


   {
   KEY  =>  'pF77',
   COMM =>  ' The name of the F77 module',
    }, 

   {
   KEY  =>  'pF77_V',
   COMM =>  ' The release of F77',
    }, 

 {
   KEY  =>  'pFilter',
   COMM =>  ' The name of the Filter module',
    }, 


  {
   KEY  =>  'pFilter_V',
   COMM =>  ' The release of Filter',
    }, 


 {
   KEY  =>  'pTBal',
   COMM =>  ' The name of the Text::Balanced module',
    }, 


  {
   KEY  =>  'pTBal_V',
   COMM =>  ' The release of Text::Balanced',
    }, 


   {
   KEY  =>  'pPDL',
   COMM =>  ' The name of the PDL module',
    }, 

   {
   KEY  =>  'pPDL_V',
   COMM =>  ' The release of PDL',
    }, 

   {
   KEY  =>  'pTk',
   COMM =>  ' The name of the TK module',
    }, 

   {
   KEY  =>  'pTk_V',
   COMM =>  ' The release of Tk',
    }, 


   {
   KEY  =>  'pPGPLOT',
   COMM =>  ' The name of the PGPLOT module',
    }, 

   {
   KEY  =>  'pPGPLOT_V',
   COMM =>  ' The release of PGPLOT',
    }, 


   {
   KEY  =>  'pCFITSIO',
   COMM =>  ' The name of the CFITSIO module',
    }, 

   {
   KEY  =>  'pCFITSIO_V',
   COMM =>  ' The release of CFITSIO',
    }, 


   {
   KEY  =>  'pTie',
   COMM =>  ' The name of the Tie::IxHash module',
    }, 

   {
   KEY  =>  'pTie_V',
   COMM =>  ' The release of Tie',
    }, 


   {
   KEY  =>  'pString',
   COMM =>  ' The name of the String::Approx module',
    }, 

   {
   KEY  =>  'pString_V',
   COMM =>  ' The release of String::Approx',
    }, 


   {
   KEY  =>  'pTerm',
   COMM =>  ' The name of the Term::Readkey module',
    }, 

   {
   KEY  =>  'pTerm_V',
   COMM =>  ' The release of Term::ReadKey',
    }, 


    {
   COMM =>  ' ==== X11 library and X11 include directories',
    }, 


   {
   KEY  =>  'X11LIB',
   COMM =>  'where are X11 libraries on your system',
    }, 

   {
   KEY  =>  'X11INC',
   COMM =>  'where are X11 include files on your system',
    }, 



    {
   COMM =>  ' ==== The pipeline modules',
    }, 

   {
   KEY  =>  'mkPIPE',
   COMM =>  ' Install the Pipeline modules [y/n] ?',
    }, 




   {
   KEY  =>  'PIPELINE',
   COMM =>  ' The name of the pipeline package',
    },

   {
   KEY  =>  'PP',
   COMM =>  'Set Y[es]/N[o] to (re)install the PP libraries ',
    },


   {
   KEY  =>  'PREREPIX',
   COMM =>  ' Set Y[es]/N[o] to install PREREPIX',
    }, 

   {
   KEY  =>  'ASTROMETRIX',
   COMM =>  ' Set Y[es]/N[o] to install ASTROMETRIX',
    }, 


       {
   KEY  =>  'PHOTOMETRIX',
   COMM =>  ' Set Y[es]/N[o] to install PHOTOMETRIX',
    }, 

);

# Here values are stored
%BuildCVals = (

INSTALL => "$ENV{HOME}/PipeLibs/",  

mkEXTERN=>'Y',
CFITSIO => 'cfitsio3210.tar.gz',
CFITSIO_LIB => '',
CFITSIO_INC => '',

PGPLOT => 'pgplot5.22.tar.gz',
PGPLOT_LIB => '',

CDSCLIENT => 'cdsclient-3.4.tar.gz',

mkPIPE =>'Y',
PIPELINE => "Pipe",
PREREPIX => 'Y',
ASTROMETRIX => 'Y',
PHOTOMETRIX => 'Y',
PP => 'Y',

mkPERL => 'Y',
pF77_V => 1.16, 
pF77 => 'ExtUtils-F77-1.16.tar.gz',

pFilter_V => 1.37, 
pFilter => 'Filter-1.37.tar.gz',

pTBal_V => 2.02, 
pTBal => 'Text-Balanced-2.02.tar.gz',


pPDL_V => '2.4.5', 
pPDL => 'PDL-2.4.5.tar.gz',

pTk_V => 804.028,
pTk => 'Tk-804.028.tar.gz',

pPGPLOT_V => 2.20,
pPGPLOT => 'PGPLOT-2.20.tar.gz',


pCFITSIO_V => 1.05,
pCFITSIO => 'Astro-FITS-CFITSIO-1.05.tar.gz',

pTie_V => 1.21, 
pTie => 'Tie-IxHash-1.21.tar.gz',

pString_V => 3.26,
pString =>  'String-Approx-3.26.tar.gz',

pTerm_V => 2.30,
pTerm => 'TermReadKey-2.30.tar.gz',

X11LIB => '',
X11INC => '',
);

}


sub mkBConf
{
open (FILE, ">$BuildConf"); 
my $entry;
foreach $entry (@BuildConf) {
  my $nokey=0;
   if (exists $entry->{KEY}) {
      my $key=$entry->{KEY};
      print FILE "$key\t";
      if (exists $BuildCVals{$key}) {
           print FILE "$BuildCVals{$key}\t";
	}  
      elsif (exists $entry->{VAL}) {
           print FILE "$entry->{VAL}\t";
	}   
    } else {$nokey=1}
    if (exists $entry->{COMM}) {
        my $comm=$entry->{COMM};
         if ($nokey==1 && $comm !~ '--') {print FILE "\t\t\t"}
	 
           print FILE "#$comm";
	}   
  print FILE "\n";	
   }
   
 close (FILE);  
   
 }


sub ReadBConf
{
  my $pfile;

  if ($#_ >= 0) {$pfile=shift}
  else {$pfile=$BuildConf}

  return if (! -e $pfile);

  open (FILE, "<$pfile"); 

  while (<FILE>) {
    chomp;

    next if ($_ =~ /^(\s*|\t*)\#/); 
    my ($vals, $comm)=split(/\#/, $_);
    my ($key, $val)=split(/\s+|\t+/, $vals);

    #  next if (! exists $BuildCVals{$key});

    $BuildCVals{$key} = $val;
  }

  close (FILE);  
}




1;

