package fitsio;

use PDL;

use Astro::FITS::CFITSIO qw( :longnames :constants);
use Tie::IxHash;
#use subs qw (rfits wfits);

sub wfits
{

my ($pdl, $file, $header, $BITPIX)=@_;

my $ret="";
my $doconv;

my $status;

if (-e $file) {system "rm $file"}

  if ($BITPIX eq "") {
 #the precision is that of the pdl array
   my $type=lc($pdl->info("%T"));
      $BITPIX =   8 if $type eq 'byte';
      $BITPIX =  16 if $type eq 'short' || $type eq 'ushort';
      $BITPIX =  32 if $type eq 'long';
      $BITPIX = -32 if $type eq 'float';
      $BITPIX = -64 if $type eq 'double';
      $doconv=-1;
   } else {$doconv=1}



  # Automatically figure output scaling

   my $bzero = 0; my $bscale = 1;
   if ($BITPIX>0) {
    #print "Warning, not writing a float (BITPIX $BITPIX)!!!\n";
      my $min = $pdl->min;
      my $max = $pdl->max;
      my ($dmin,$dmax) ;
       ($dmin,$dmax) = (0, 2**8-1)     if $BITPIX == 8;
      ($dmin,$dmax) = (-2**15, 2**15-1)  if $BITPIX == 16; #STORE as SHORT
      ($dmin,$dmax) = (-2**31, 2**31-1)  if $BITPIX == 32;

      if ($min<$dmin || $max>$dmax) {
         $bzero = $min;
         $max -= $bzero;
         $bscale = $max/$dmax if $max>$dmax;
       }
       $$header{BSCALE}=$bscale; $$header{BZERO}=$bzero;
       $pdl -= $bzero;
       $pdl /= $bscale; 
   }

if ($doconv == 1) {
  my $convert = sub { return $_[0] };# Default - do nothing
   $convert = sub {byte($_[0])}   if $BITPIX ==   8;
   $convert = sub {short($_[0])}  if $BITPIX ==  16;
   $convert = sub {long($_[0])}   if $BITPIX ==  32;
   $convert = sub {float($_[0])}  if $BITPIX == -32;
   $convert = sub {double($_[0])} if $BITPIX == -64;
    $pdl = &$convert($pdl); 
}

   #update dimensions and precision
   my ($ndims, $k);

   $ndims = $pdl->getndims; # Dimensions of data array 
   
   #$ndims=2;   #Only 2d images are written for the moment
   $$header{NAXIS}  = $ndims;
   for $k (1..$ndims) { $$header{"NAXIS$k"} = $pdl->getdim($k-1) }
   $$header{BITPIX}=$BITPIX;

$status=0;

#open the file
my $fptr = Astro::FITS::CFITSIO::create_file($file,$status);
if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); return $ret}
 
my ($naxis1, $naxis2);
	
$naxis1=$$header{NAXIS1};
if (exists $$header{NAXIS2}) {
   $naxis2=$$header{NAXIS2}
  } 

#Create the header
$fptr->write_key(TLOGICAL,"SIMPLE",1,"Standard FITS",$status);
if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); $fptr->close_file($status);return $ret}

#$fptr->write_key(TSHORT,"BITPIX",$BITPIX,"Number of bits per data pixel",$status);
#if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); $fptr->close_file($status);return $ret}

# ////////////////////////////////////////////////////////////////
# THE IMAGE HEADER COULD BE ALSO INITIALIZED AS FOLLOWS  
# my $naxis=2;
# my $naxes=[$naxis1,$naxis2];
# fits_write_imghdr($fptr,$BITPIX,$naxis,$naxes,$status);
# delete $$header{BITPIX};
# delete $$header{NAXIS};
# delete $$header{NAXIS1};
# delete $$header{NAXIS2};
# //////////////////////////////////////////////////////////////


#write the header
    my @keys=keys %$header ;
 foreach my $i (1..$#keys){
         my $h=$keys[$i];
	 my $val=$$header{$h};
	 my $comm= $$header{COMMENTS}{$h};
	 my $type=getheadtype(\$val);
	 if ($h =~ /COMMENT/) {Astro::FITS::CFITSIO::fits_write_comment($fptr,$comm,$status)}
	 elsif  ($h =~ /HISTORY/) {Astro::FITS::CFITSIO::fits_write_history($fptr,$comm,$status)}
	 else {$fptr->write_key($type,$h,$val,$comm, $status)}
         if ($status ne 0) {
	   Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); 
	   $fptr->close_file($status);
	   system ("rm $file") if (-e $file);
	   return $ret
	 }
   }
   
#warn "whead returns $status!\n";

#OK, write the image
$status=0;


my %funcs = (
        '8' => { 'cfitsio' => \&Astro::FITS::CFITSIO::fits_write_2d_byt },
        '16' => { 'cfitsio' => \&Astro::FITS::CFITSIO::fits_write_2d_sht },
        '32' => { 'cfitsio' => \&Astro::FITS::CFITSIO::fits_write_2d_lng },
        '-32' => {'cfitsio' => \&Astro::FITS::CFITSIO::fits_write_2d_flt },
        '-64' =>  { 'cfitsio' => \&Astro::FITS::CFITSIO::fits_write_2d_dbl },
);

if (exists $funcs{$BITPIX})
{
    &{$funcs{$BITPIX}{'cfitsio'}}($fptr,1,$naxis1,$naxis1,$naxis2,$pdl->get_dataref,$status);
   if ($status ne 0) {
                Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); 
                $fptr->close_file($status);                
	        system ("rm $file") if (-e $file);
		return $ret;
              }
    $fptr->close_file($status);
}
else
{
    $fptr->close_file($status);
    $ret= "$0: invalid BITPIX keyword (= $BITPIX) in image '$file'";
}

warn $ret if $ret;

   return $ret;

} #end sub wfits




sub rhead {

#store the header of a FITS file into an ORDERED hash
#a sequential number is appended to the HISTORY and COMMENT rows in order to store them 
#in the hash as well

my $file=shift;

my ($n, $left);

my $status = 0;
my $ret;
my $fitsfile = Astro::FITS::CFITSIO::open_file($file, READONLY, $status);
if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); return}


#Preserve the order in the hash !!!
my %header;
tie ( %header, 'Tie::IxHash');
$fitsfile->get_hdrspace($n, $left, $status);

my ($key, $value, $comment);

my $nhist=1;
my $ncomm=1;

for my $i (1..$n) {

   last unless $status == 0;
   $fitsfile->read_keyn($i, $key, $value, $comment, $status);
   next if  ($key =~ /BSCALE/i || $key =~ /BZERO/i);
   if ($key =~ /HISTORY/) {
      $key=$key.$nhist;    #this will create a key HISTORY1 ...
      $nhist++;             
     } 
   elsif ($key =~ /COMMENT/)  { 
       $key=$key.$ncomm;        #this will create a key COMMENT1 ...
       $ncomm++
    } 
       $header{$key} = $value;
       $header{COMMENTS}{$key} = $comment;          
   } 
   
   
$fitsfile->close_file($status);
return %header;
} #end sub rhead


sub getheadtype
{
#get the right type for the values in the header (called by wfits)
#the code was borrowed from the PDL IO library

my $val=shift;

my $typ;

 if ($$val =~ /^ *([+-]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))? *$/) { # Number?
        $typ=TFLOAT;
	$typ=TSHORT if ($$val !~ /\./) ;	#Is it really a float ?
  }  elsif ($$val eq 'F' or $val eq 'T') {          # Logical flags ?
      $typ=TLOGICAL;
      
} else {
         $$val=$1 if ($$val=~m/'(.*)'/); # Take off surrounding ''
          $typ=TSTRING;
  }
  
 return $typ;
}  #end sub getheadtype	  

sub rfits
{
Astro::FITS::CFITSIO::PerlyUnpacking(0);

my ($file, $ncols, $nrows, $type)= @_;

if (! $ncols) {$ncols=-1};
if (! $nrows) {$nrows=-1};
if (! $type) {$type = 'float'}


my $ret="";

my $status = 0;
my $fptr = Astro::FITS::CFITSIO::open_file($file, READONLY, $status);

if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); return (null,$ret)}

my ($naxis1, $naxis2, $bitpix, $bscale, $bzero);

$fptr->read_key(TLONG,'BITPIX',$bitpix,undef,$status);
if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret);
                    $fptr->close_file($status);return (null, $ret)
		   }

if ($ncols <=0) { 
   $fptr->read_key(TSTRING,'NAXIS1',$naxis1,undef,$status);
   if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); 
                       $fptr->close_file($status);return (null, $ret)
		    }
} else {$naxis1=$ncols}

if ($nrows <=0) { 
   $fptr->read_key(TSTRING,'NAXIS2',$naxis2,undef,$status);
   if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); 
              $fptr->close_file($status);return (null, $ret)
	 }
} else {$naxis2=$nrows}

$fptr->read_key(TLONG,'BSCALE',$bscale,undef,$status);
$bscale="" if ($status ne 0);
$fptr->read_key(TLONG,'BZERO',$bzero,undef,$status);
$bzero="" if ($status ne 0);

$status=0;

my %funcs = (
        '8' => { 'pdl' => \&byte, 'cfitsio' => \&Astro::FITS::CFITSIO::fits_read_2d_byt },
        '16' => { 'pdl' => \&short , 'cfitsio' => \&Astro::FITS::CFITSIO::fits_read_2d_sht },
        '32' => { 'pdl' => \&long, 'cfitsio' => \&Astro::FITS::CFITSIO::fits_read_2d_lng },
        '-32' => { 'pdl' => \&float , 'cfitsio' => \&Astro::FITS::CFITSIO::fits_read_2d_flt },
        '-64' => { 'pdl' => \&double , 'cfitsio' => \&Astro::FITS::CFITSIO::fits_read_2d_dbl },
);

#Handle unsigned short

if ($bscale ne "" && $bzero ne "" && $bitpix==16) {
 # if a section was given, the image must be read as short and then rescaled
 # (a bug in the Astro::FITS::CFITSIO library ????)
 
  if ($ncols<=0 && $nrows<=0) { 
     $funcs{$bitpix}{'pdl'}=\&ushort;
     $funcs{$bitpix}{'cfitsio'}=\&Astro::FITS::CFITSIO::fits_read_2d_usht;
     $bscale="";
     $bzero="";
  }
}

my ($pdl,$anynul);
if (exists $funcs{$bitpix})
{
    $pdl = &{$funcs{$bitpix}{'pdl'}} (zeroes($naxis1,$naxis2));
    &{$funcs{$bitpix}{'cfitsio'}}($fptr,1,0,$naxis1,$naxis1,$naxis2,${$pdl->get_dataref},$anynul,$status);    
    if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); $fptr->close_file($status);return (null,$ret)}
    $pdl->upd_data;
    $fptr->close_file($status);
    if ($type eq 'byte') { $pdl=byte($pdl)}    
    elsif ($type eq 'short') { $pdl=short($pdl)}    
    else {$pdl=float($pdl)}; #convert piddle to float
    #print ">>>> $file: ", $pdl->info("%T"),"\n";
    if ($bscale ne "" && $bzero ne "") { $pdl *= $bscale; $pdl += $bzero} #see above
  }
else
{
    $fptr->close_file($status);
    $ret= "$0: invalid BITPIX keyword (= $bitpix) in image '$file'";
}
#print "av ", average($pdl->clump(2)),"\n";

Astro::FITS::CFITSIO::PerlyUnpacking(1);
	
 return ($pdl, $ret);

} #end sub rfits

sub whead
{
 my ($file, $header)=@_;
 my $status = 0;
 my $ret;
 my $fptr = Astro::FITS::CFITSIO::open_file($file, READWRITE, $status);

if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret);   die $ret}
   my @keys=keys %$header ;



 foreach my $i (0..$#keys){
         my $h=$keys[$i];
	 next if $h eq 'SIMPLE';
	 my $val=$$header{$h};
	 my $comm= $$header{COMMENTS}{$h};
	 my $type=getheadtype(\$val);
	 if ($h =~ /COMMENT/) {Astro::FITS::CFITSIO::fits_write_comment($fptr,$comm,$status)}
	 elsif  ($h =~ /HISTORY/) {Astro::FITS::CFITSIO::fits_write_history($fptr,$comm,$status)}
	 else {$fptr->update_key($type,$h,$val,$comm, $status)}
         if ($status ne 0) {Astro::FITS::CFITSIO::fits_get_errstatus($status,$ret); $fptr->close_file($status);return $ret}
   }
   $status = 0;
   $fptr->close_file($status);

   
   return $ret;
}


-1;
