package Header;

use Carp;
use vars qw($AUTOLOAD);

use  Astro::FITS::CFITSIO qw( :longnames :constants);
use Common::fitsio;

use strict;

my @dbKeys=qw/ORIGIN DETECTOR RUNID OBJECT FILENAME DATE LST-OBS LOCDATE LOCTIME RA DEC OBSTYPE EXPTIME FILTER/;



   sub AUTOLOAD {
               my $self = shift;
               my $type = ref($self)
                           or croak "$self is not an object";

               my $name = $AUTOLOAD;
               $name =~ s/.*://;   # strip fully-qualified portion
         
               if (@_) {
                   return $self->{$name} = shift;
               } else {

                   return if (! exists  $self->{$name});
                   my $val= $self->{$name};
		   		   $val=~tr/\/\'| //d;
		   if ($val eq '') {$val='X'}
		   elsif ($name eq 'EXPTIME') {
		        $val=int($val+.5);
		     }
		   elsif ($name eq 'OBJECT') {
                  $val=~ tr/,/_/;            # translate comma into underscore
                  $val=~ tr/0-9a-zA-Z_//cd;   # get rid of all "non word" characters
                       }  
		   return $val;
               }
           }

sub new {

 my $proto = shift;
 my $class = ref($proto) || $proto;

my $self = {
  ORIGIN => undef,
 INAME => undef,
 DETECTOR  => undef,
 RUNID=> undef,
 RA => undef,
 DEC => undef,
 OBSTYPE => undef,
 EXPTIME => undef,
 FILTER => undef,
 DATE => undef,
 LOCTIME => undef,
 LOCDATE  => undef,
 'LST-OBS' => undef, 
 OBJECT  => undef,
 IMAGEID  => undef,
 FILENAME => undef,
 #AM => undef,
 #LAT => undef,
 _table => {}   ,  #the conversion table for keywords
 _properties => {   #the default CCD properties
   MINCCD=> undef,
   NCCDX=> undef,
   NCCDY => undef,
   NUMCCD=>undef,
   data => {
       section => []
    },   
   trim => {
     section =>[]
   },
   overscan => {
     section=>[],
     order => undef
   },  
   FITS_TYPE => undef         #(SIMPLE/MEF)
  }  #end of _properties
 }  ;

 bless ($self, $class);

return ($self)

} #end new


sub getVals
{
#read the header of FITS file
my %head;


 my $self=shift;
 my $input=shift;
 

 if (ref($input) eq 'HASH') {%head=%$input}
 else {
      if ($input !~ /fits$/) {$input = $input.'.fits';}
      %head = &fitsio::rhead($input);
    }
 

#print " ??? ", $self->{FILENAME}, "\n";

 
  foreach  my $key (keys %$self) {
  
      next if (ref($self->{$key}) eq 'HASH' || ! exists $self->{_table}{$key});
      my $kid=$self->{_table}{$key};
      next if (! exists $head{$kid});
      my $value=$head{$kid};
      $value =~ tr/ //d;
      $self->{$key}=$value;
   }

 my @s=split(/\//, $input );
 $self->{INAME}=$s[-1];

#print " ??? ", $self->{FILENAME}, "\n";

return %head;

} #end sub getValues


sub sexRA {
#Convert RA from decimal degrees to hh:mm:ss.s
my $self=shift;
my $tra=shift;



  $tra/=15.;  #ra is in degrees
  my $ra1=int($tra);
  my $ra3=($tra-$ra1)*60.;
  my $ra2=int($ra3);
  $ra3=($ra3-$ra2)*60;
  my $ira3=int($ra3);
  my $RA=sprintf ("%02d:%02d:%02d.%.0f",$ra1,$ra2,$ira3,100*($ra3-$ira3));

#warn "input $tra -> $ra1 $ra2 $ra3 $ira3 -- $RA";

   return $RA;

}   #end sub sexRA

sub sexDEC {
#Convert DEC from decimal degrees to dd:mm:ss.s
my $self=shift;
my $tdec=shift;
return 0 if ($tdec==0);


  my $sig=$tdec/abs($tdec);
  my $dec1=int($tdec);
  #if ($dec1 < 0 && ($d-$dec1) !=0) {$dec1--}
  my $dec3=abs($tdec-$dec1)*60.;
  my $dec2=int($dec3);
  $dec3=($dec3-$dec2)*60;
  my $idec3=int($dec3);
  
   my $DEC=sprintf ("%02d:%02d:%02d.%.0f",$dec1,$dec2,$idec3,100*($dec3-$idec3));
   
  if ($sig==-1 && $dec1 ==0)  {$DEC='-'.$DEC}
  elsif ($dec1 !~ /^-/ && $dec1 >=0) {$DEC='+'.$DEC}
  
   return $DEC;
   
}  #ens sub sexDEC

sub set_locdate {
# this subroutine computes the "local" date
# it shift back the date of 1 day if LT < 12
# the date is presented and returned as : YYYY-MM-DD

my $self=shift;

  my @loctime=split (/:/, $self->{LOCTIME}, 2);
  my $locdate=$self->{LOCDATE};
 

  if ($loctime[0] lt 12) {
     $locdate=$self->yesterday($locdate);
   }

#warn "$locdate @loctime";   

 $self->{LOCDATE} = $locdate;

}  #end sub locdate

sub UT2Local 
  {
    my $self=shift;

    my ($date, $ut, $tzone)= @_;
    my ($locdate, $loctime);
    #my $tzone=4;         #----------- the TIMEZONE (must be SUBTRACTED from UT)
    my @HH=split(/:/,$ut,3);
    if ($HH[0] < $tzone) {$locdate=$self->yesterday($date); $HH[0] += 24}
    elsif ($HH[0]-$tzone > 24) {$locdate=$self->tomorrow($date); $HH[0]-=24}
    else {$locdate=$date};
    $HH[0] -= $tzone;
       $loctime=join(':',@HH);
#warn "???????????/ $date - $ut - $locdate - $loctime" if $HH[0]<0;
    return ($locdate, $loctime);
  }



sub yesterday {
  
  my $self=shift;

  my $odate=shift;
  my $ndate;
  
  my ($YEAR,$MONTH,$DAY)=split(/-/,$odate); #
  my ($newDAY, $newMONTH, $newYEAR);
  if ($DAY gt "01")		# this is NOT the first day of the month : only DAY is shifted
    {$newDAY=$DAY-1;
     $newDAY=sprintf "%02d",$newDAY; # make sure to get 2 digits format
     $ndate=join('-',$YEAR,  $MONTH, $newDAY); # concatenate to YYYY-MM-DD
     #$ndate=$YEAR ."-". $MONTH ."-". $newDAY;
   }
  else {			# this is the first day of the month, so DAY and MONTH are shifted
    if ($MONTH eq "01")
      {				# HEY ! this is New Year's Day  !!! then shift also YEAR
	$newMONTH="12";
	$newDAY="31";
	$newYEAR=$YEAR-1;
        $ndate=join('-',$newYEAR,  $newMONTH, $newDAY);
	#$ndate=$newYEAR ."-". $newMONTH ."-". $newDAY; # concatenate to YYYY-MM-DD
      }
    else {			# this is just a first day of other month than January,
      #             so shift the MONTH and DAY ...
      $newMONTH=$MONTH -1;
      $newMONTH=sprintf "%02d",$newMONTH; # 2 digits format
      #  ... and then shift the day to the last day of the previous month ;
      #  to calculate the last day of the previous, use an hash array
      #    built as a serie of couples : first couple is "02","31"  and so on...
      #    this must be understood as : "the last day of the previous month is that"
      #    for instance : the current month is February(02), then the
      #                     previous month is January, then last is 31
      
      my (%jour);
      
      my $flag=0;
      if ($YEAR == 2000) {$flag=1} else {
	my $Y4=$YEAR/4.;
	if (int($Y4) == $Y4) {$flag=1}
      }
      
      # for each month, give the days of the month BEFORE
      if ($flag==1 ){		# leap year - bissextile
	(%jour)= ("02","31","03","29","04","31","05","30","06","31","07","30","08","31","09","31","10","30","11","31","12","30");
      }
      else {
	(%jour)= ("02","31","03","28","04","31","05","30","06","31","07","30","08","31","09","31","10","30","11","31","12","30");
      }
      $ndate=join('-',$YEAR,  $newMONTH, $jour{$MONTH});
      #$ndate=$YEAR ."-". $newMONTH ."-". $jour{$MONTH};
     }
   }
 return $ndate;
 
} #end sub yesterday

sub tomorrow {

  my $self=shift;
  
  my $odate=shift;
  my $ndate;
  
  my ($YEAR,$MONTH,$DAY)=split(/-/,$odate); #
  
  my (%jour);
  
  my $flag=0;
  if ($YEAR == 2000) {$flag=1} else {
    my $Y4=$YEAR/4.;
    if (int($Y4) == $Y4) {$flag=1}
  }
  
  # for each month, give its days
  if ($flag==1 ){		# leap year - bissextile
    (%jour)= ("01","31","02","29","03","31","04","30","05","31","06","30","07","31","08","31","09","30","10","31","11","30","12","31");
  }
  else {
    (%jour)= ("01","31","02","28","03","31","04","30","05","31","06","30","07","31","08","31","09","30","10","31","11","30","12","31"); 
  }
  
  if ($DAY lt $jour{$MONTH})	# this is NOT the last day of the month : only DAY is shifted
    {$DAY++;
     $DAY=sprintf "%02d",$DAY;	# make sure to get 2 digits format
   }
  else {			# this is the last day of the month, so DAY and MONTH are shifted
    if ($MONTH eq "12")
      {				# HEY ! this is New Year's Day  !!! then shift also YEAR
	$MONTH="01";
	$DAY="01";
	$YEAR++;   
      }
    else {			# this is just a last day of other month than December,
      #             so shift the MONTH and DAY ...
      $MONTH++;
      $DAY='01';
      $MONTH=sprintf "%02d",$MONTH; # 2 digits format
      #  ... and then shift the day to the last day of the previous month ;
      #  to calculate the last day of the previous, use an hash array
      #    built as a serie of couples : first couple is "02","31"  and so on...
      #    this must be understood as : "the last day of the previous month is that"
      #    for instance : the current month is February(02), then the
      #                     previous month is January, then last is 31
      
    }  
  }
  
  $ndate= join('-', $YEAR, $MONTH, $DAY) ; # concatenate to YYYY-MM-DD
  return $ndate;
  
}				#end sub tomorrow



sub mkSummary {
my $self=shift;

my $delimit=' # ';
$delimit=shift if ($#_!=-1);

my @vals=();

#print " 2 ", $self->{INAME}, "\n";

 foreach  my $key (@dbKeys) {
       my $value;
       if (exists $self->{$key} && defined $self->{$key})  {
#            eval("\$value=\$self->$key");;
        $value= $self->{$key};
	   
       }  else {$value='UNDEFINED'}

       if ($key eq 'FILENAME' && $value eq 'UNDEFINED') {
	       $value=$self->{INAME};
       }


       push @vals, $value;




   }


my $row=join $delimit, @vals;



return $row;

} #end sub mkSummary

sub dbKeys
{
 return @dbKeys;
}


sub Section {
my $self=shift;
my %user=@_;

# NOTE: this function is basically a copy of that in Config.pm 
#      (just the fields are changed)


#at least the step is required !
return if (! exists $user{step} );

my $step=($user{step});

 my $rec=$self->{_properties}{$step}{section};


if ( exists $user{array}) {
#0. an array is given, so just feed this into 
  my $allsec=$user{array};
    my @sections=();  #empty the array
  foreach my $i (0..$#{$allsec}) {
      my $sec={};
      $sec->{c1}= $allsec->[$i][0];
      $sec->{c2}= $allsec->[$i][1];
      $sec->{r1}= $allsec->[$i][2];
      $sec->{r2}= $allsec->[$i][3];
      push  @sections, $sec;
   }
#ok, now store the array
  $self->{_properties}{$step}{section}=[@sections];   
}

elsif (exists $user{ccd} || $step eq 'data') {
# 1> the CCD id was given

 my $ccd;

 if ($step eq 'data') {$ccd=0}
 else {$ccd=$user{ccd}}

 if (exists $user{c1} && exists $user{c2} && exists $user{r1} && exists $user{r2})
 {

  my @sections=(); 
# 1a> The section was given, so it must be modified  
  my $set=0;
      my $sec={};
      $sec->{c1}= $user{c1};
      $sec->{c2}= $user{c2};
      $sec->{r1}= $user{r1};
      $sec->{r2}= $user{r2};

#if the array contains something, then try to substitute the entry with that ccd
   $rec->[$ccd]=$sec;       
 }



#1b> return the section as an array
   my @osec=();

   my $rec=$self->{_properties}{$step}{section};
   my $xsec=$rec->[$ccd];
   @osec=($xsec->{c1}, $xsec->{c2}, $xsec->{r1}, $xsec->{r2});

 return @osec;
 }
else {
# 2> return the whole stuff (CCDs and sections)
   my @allsec=();
   foreach my $sec (@{$self->{_properties}{$step}{section}}) {
     my @sec=($sec->{c1}, $sec->{c2}, $sec->{r1}, $sec->{r2});
     push @allsec, [@sec];
   }

 return @allsec;
 } 
 
} #end sub Section


sub getProps {
my $self=shift;

my $field='';
if (@_) {$field=shift};

my $sfield='';
if (@_) {$sfield=shift};

my $val='';

if ($sfield eq 'section') {
  return
}


if ($sfield eq '') {
 if (exists $self->{_properties}{$field}) {$val=$self->{_properties}{$field} }
} else {
 if (exists $self->{_properties}{$field}{$sfield}) {$val=$self->{_properties}{$field}{$sfield} }
}


return $val;

} #end sub getProps


sub buildName 
{

my $self=shift;

my $flags='';
my $opt;
$opt = pop @_ if ref($_[$#_]) eq 'HASH'; 


if (@_) {
 my $hdr=shift;
# if (ref($hdr) eq 'HASH') {    
  foreach  my $key (keys %$self) {  
      next if (ref($self->{$key}) eq 'HASH');
      next if (! exists $hdr->{$key});
      my $value=$hdr->{$key};
      $value =~ tr/ //d;
      $self->{$key}=$value;

   }
# } else {
#    ($DETECTOR, $RUNID, $FILTER, $IMAGEID)=split(\/.\, $hdr);
#   } 
}

my $ext='fits';
if (exists $opt->{EXT}) {$ext= $opt->{EXT}}

my $name='';
if (exists $opt->{STEP}) {
   my $step=$opt->{STEP};
   if ($step eq 'bias') {$name='Bias'}
   elsif ($step eq 'dark') {$name='Dark'}
   elsif ($step eq 'mask') {$name='Mask'}
   elsif ($step eq 'dflat') {$name='DomeFlat'}
   elsif ($step eq 'tflat') {$name='TwFlat'}
   elsif ($step eq 'sflat') {$name='SuperFlat'}
   elsif ($step eq 'fringe') {$name='Fringe'}
}
elsif (exists $opt->{NAME}) {
   $name=$opt->{NAME}
}



if (exists $opt->{FLAGS}) {
   $flags=$opt->{FLAGS};
 if ($flags eq 'mos' || $flags eq 'sfrg' ||$flags eq 'objs' ) {
 #Simply replace the flag..... but this part is still in building...
     my @s=split(/\./, $name);
     splice @s, $#s-1, 1 if ($flags eq 'mos'); #remove the CCD
     if ($#s>3) {
          $s[3]=$flags;  #replace the flags
     } else {splice (@s, $#s, 0, $flags)}	  #no flag, insert it
     $name=join '.', @s;
     return $name;
 }
}

my $odir='';
if (exists $opt->{OUTDIR}) {$odir=$opt->{OUTDIR}}

my $date=0;
if (exists $opt->{DATE}) {$date=1 if lc($opt->{DATE}) eq 'yes' }

# Ok, let's build the name

#read keywords from the header and remove ' and spaces

my @hkeys=('DETECTOR', 'RUNID', 'LOCDATE', 'FILTER', 'IMAGEID');
my ($DETECTOR, $RUNID, $LOCDATE, $FILTER, $IMAGEID);


foreach my $hkey (@hkeys) { 

   if ($hkey eq 'FILTER' && ($name eq 'Bias' || $name eq 'Dark')) {
       $FILTER='';
       next;
   }
   elsif ($hkey eq 'LOCDATE' && $date==0) {next}
   
   #We want to assign to a variable called eg. $FILTER the content
   # of $self->{FILTER}
   if (!exists $opt->{$hkey}) {eval("\$$hkey=\$self->$hkey");}; 
}

$IMAGEID=sprintf("%02d",$IMAGEID);

my @name=($odir.$name, $DETECTOR, $FILTER);   #the compulsory fields


# The 'optional' fields
if ($date) {push @name, $LOCDATE}
if ($flags) {push @name, $flags}
push @name, ("#$IMAGEID", $ext);

# Make the name
$name=join '.', @name;

return $name;

}

sub Update
{
my $self=shift;
my $head=shift;

 while  (my ($key, $value) = each %{$self}) {
    next if ref($value) eq 'HASH'; 
    my $hval=$head->{$key};
    $hval =~ tr/\'| //d;
     next if $value eq $hval;
    my $bkey="ORIGINAL $key"; 
    if (exists $head->{$key}) {
       my $bkey="HISTORY$key"; 
       $head->{$bkey}=' ';
       my $h=$key.' = '.$head->{$key}.' Modified by PREREPIX';
       $head->{COMMENTS}{$bkey}=$h; 
    }  else {
        $head->{COMMENTS}{$key}='Added by PREREPIX';
    }  
    $head->{$key}=$value;       
 }
}


#sub setAirMass
#  {
#  #From IRAF noao$astutil/airmass.x
#  my $self=shift;

 
#  my $ha=sex2dec($self->{'LST-OBS'})-sex2dec($self->{RA});
#   $ha+=24. if ($ha<0);
# local $scale;
#  *scale=\750.0;  #Scale factor of the Earth's atmosphere.
# my $lat=sex2dec($self->{'LAT'});
# my $dec=sex2dec($self->{'DEC'});

#  my cos_zd = sin(DEGTORAD($lat)) * sin(DEGTORAD($dec)) +
#                 cos(DEGTORAD($lat)) * cos(DEGTORAD($dec)) * cos(DEGTORAD($ha*15.))

#   my $x=$scale*$cos_zd;

#  return (sqrt ($x**2 + 2*$scale + 1) - $x);

#  }

#sub DEGTORAD
#  {
#  my $deg=shift;
# local $PI;
# *PI=\3.14159265358979;
#  my $rad=$deg*$PI/180.;
#  return $rad;
#  }

#sub sex2dec
#  {
#  my $sex=shift;
#  my @s=split(/:/, $sex,3);
#  my $dec=$s[0]+$s[1]/60.+$s[2]/3600.;
#  return $dec;
#  }

1;
