package IConfig;
$VERSION = 1.00;
use strict;

{
my @_Configs;

#my $lsteps=qw('Bias' 'Dark' 'DFlat'  'TFlat'  'SFlat'  'Fringe' 'Science');

sub new  {


my $tn='';

my $class=shift;

if (@_) {$tn=shift} else {
  my $nc=$#_Configs+2;
  $tn=sprintf("%s%02d.par",$main::CONFIG,$nc);
}

my $params= {
  NAME => $tn,
  TOSTEP => 'science',
  ENTRIES => [],         #used internally

  #INSTRUMENT => '',
  NUMCCD => '',    # the total number of CCDs
  MINCCD => 0,     # the starting CCD index (should be the lowest value in CCDid....)
  NCCDX => '',     # how many CCDs per row
  NCCDY => '',     # how many CCDs per column
  CCDid => [],     # the id of each CCD (should be a number)
  CCDpos => [],     # the extension of each CCD (only meaningful for MEF files)

  steps=> [qw(general select overscan trim mask mosaic bias dark dflat tflat fringe sflat science)],

  general => {
    fields => [qw(instrument fits_type sections)],  #tie the options to the given order
      instrument => {Value=> 'CFH12K', Allowed=>$main::IMPL_INSTR, Help=>'The instrument used'},
      fits_type  => {Value=> 'SIMPLE', Allowed=>['SIMPLE','MEF'], Help=>'The format of the FITS files'},
      sections => {Help=>'Define the overscan/trim sections'},
      section => [],
   } ,

  select => {
   main => 'general',  #this means that this field will be printed inside that box
    fields => [qw(xfilter filter xsatur highrej)],
      xfilter => {Value=> 'no', Allowed=>['no', 'is', 'contains', 'begins_with'], Help=>'Extract only images for a given filter ?'},
      filter => {Value=> '', Help=>'The name of the filter'}, 
      xsatur => {Value=> 'yes', Allowed=>['yes', 'no'], Help=>'Skip images with average higher than the saturation limit when combining ?'}, 
      highrej => { Value=> '50000', Help=>'The saturation limit in counts'},
  }, 

  overscan => {
   main => 'general',  #this means that this field will be printed inside that box
    fields => [qw(apply method order mode plot)],
     apply => {Value=> 'no', Allowed=>['yes','no'], Help=>'Apply overscan strip correction?'},
     section => [],
     method => {Value=> 'fit', Allowed=>['fit','block'], Help=>''},
     order =>{Value=> "", Help=>'Order of the polynomial fitting the overscan OR size of the block'},
     mode =>{Value=> 'median', Allowed=>['median','average'], Help=>'How to average the os section'},
     plot => {Value=> 'no', Allowed=>['yes','no'], Help=>'Make a plot of the overscan'}
  },

  trim => {
   main => 'general',
   fields => [qw(apply)],
     apply => {Value=> 'no', Allowed=>['yes','no'], Help=>'Trim the image ?'},
     section => [],
  },

  mask => {
  main => 'general',
   fields => [qw(use input output badvalue)],
    use => { Value=> 'no', Allowed=>['yes','no'], Help=>'Use bad pixel masks ?'},
    input => { Value=> '', Help=>'List of the bad pixel mask files'},
    output => { Value=> '', Help=>'Output bad pixel mask files'},
    badvalue => { Value=> '-1000', Help=>'Value assigned to bad pixels'},
#    highrej => { Value=> '60000', Help=>'Upper value to reject'},
  },
 
  mosaic => {
   main => 'general',
   fields => [qw(binx biny sigma iterations)],
    binx => { Value=> 10, Help=>'Binning in X'},
    biny => { Value=> 10, Help=>'Binning in Y'},
    sigma => {Value=>2, Help=>'Background estimate: Sigma rejection'},  
    iterations => {Value=>1, Help=>'Background estimate: Number of iterations performed'},
  },
 

  bias => {
  fields => [qw(apply input output combine scale sigma niter mkmosaic)],
   apply=> { Value=> 'no', Allowed=>['yes','no'], Help=>'Compute/subtract bias ?'},
   combine=> { Value=> 'avsig', Allowed=>['average','median','avsig'], Help=>'Combine method'},
   scale=> { Value=> 'none', Allowed=>['none','median','exposure'], Help=>'Scaling method'},
   sigma=> { Value=> '2', Help=>'Sigma rejection if combine=avsig'},
   niter=> { Value=> '3', Help=>'Number of iterations if combine=avsig'},
   input=> { Value=> '', Help=>'Input list of bias frames'},
   output=> { Value=> 'Pbias.lst', Help=>'Output list of bias frames'},
   mkmosaic=>{ Value=> 'yes', Allowed=>['yes','no'], Help=>'Produce a mosaic ? '},
   done => 0,
   CHECK => ''
  },

  dark => {
  fields => [qw(apply input output combine scale  sigma niter mkmosaic)],
   apply=> { Value=> 'no', Allowed=>['yes','no'], Help=>'Compute/subtract dark ?'},
   combine=> { Value=> 'avsig', Allowed=>['average','median','avsig'], Help=>''},
   scale=> { Value=> 'exposure', Allowed=>['none','median','exposure'], Help=>''},
   sigma=> { Value=> '2', Help=>'Sigma rejection if combine=avsig'},
   niter=> { Value=> '3', Help=>'Number of iterations if combine=avsig'},
   input=> { Value=> '', Help=>'Input list of dark frames'},
   output=> { Value=> 'Pdark.lst', Help=>'Output list of bias frames'},
   mkmosaic=>{ Value=> 'yes', Allowed=>['yes','no'], Help=>'Produce a mosaic ? '},
    done => 0 ,
   CHECK => ''
  },

  dflat => {   # dome flat correction 
  fields => [qw(apply input output combine scale  sigma niter normalize nmode 
     mkmosaic)],
   apply=> { Value=> 'no', Allowed=>['yes','no'], Help=>'Compute/apply a dome flat ?'},
   combine=> { Value=> 'avsig', Allowed=>['average','median','avsig'], Help=>''},
   scale => { Value=> 'median', Allowed=>['none','median','exposure'], Help=>''},
   sigma=> { Value=> '2', Help=>'Sigma rejection if combine=avsig'},
   niter=> { Value=> '3', Help=>'Number of iterations if combine=avsig'},
   normalize=> { Value=> 'yes', Allowed=>['yes','no'], Help=>'Normalize after combination ?'},
   nmode=> { Value=> 'global', Allowed=>['global','single'],
           Help=>'Normalize over the whole mosaic [g] or on single frames [s]'},
   input => { Value=> '', Help=>'Input list of dome flat frames'},
   output => { Value=> 'Pdflat.lst', Help=>'Output list of dome flat frames'},
   mkmosaic=>{ Value=> 'yes', Allowed=>['yes','no'], Help=>'Produce a mosaic ? '},
   done => 0,
   CHECK => ''
 },

  tflat => {   # twilight flat correction 
  fields => [qw(apply input output combine scale  sigma niter normalize nmode 
     mkmosaic)],
   apply=> { Value=> 'no', Allowed=>['yes','no'], Help=>'Compute/apply a twilight flat ?'},
   combine=> { Value=> 'avsig', Allowed=>['average','median','avsig'], Help=>''},
   scale => { Value=> 'median', Allowed=>['none','median','exposure'], Help=>''},
   sigma=> { Value=> '2', Help=>'Sigma rejection if combine=avsig'},
   niter=> { Value=> '3', Help=>'Number of iterations if combine=avsig'},
   normalize=> { Value=> 'yes', Allowed=>['yes','no'], Help=>'Normalize after combination ?'},
   nmode=> { Value=> 'global', Allowed=>['global','single'],
           Help=>'Normalize over the whole mosaic [g] or on single frames [s]'},
   input => { Value=> '', Help=>'Input list of twilight flat frames'},
   output => { Value=> 'Ptflat.lst', Help=>'Output list of twilight flat frames'},
   mkmosaic=>{ Value=> 'yes', Allowed=>['yes','no'], Help=>'Produce a mosaic ? '},
   done => 0,
   CHECK => ''
 },

 sflat => {    # super flat 
 fields => [qw(apply input output preproc backsize group combine scale  sigma niter 
            normalize nmode mkmosaic)],
   apply => { Value=> 'no', Allowed=>['yes','no'], Help=>'Compute/apply a super flat ?'},
   input => { Value=> '', Help=>'Input list of the images used to compute the super flat frames'},
   output => { Value=> 'Psflat.lst', Help=>'Output list of super flat frames'},
# Preprocessing of the frames
   preproc => { Value=> 'SEx', Allowed=>['SEx','smooth','none'],
           Help=>'Extract background with SExtractor BEFORE combining [SEx], AFTER combining [smooth] or simply combine [none] ?'},
   backsize => { Value=> '300', Help=>'Background mesh size if SExtractor is used'},
# Combination of the frames into a master frame
  group  => { Value=> 'bydate', Allowed=>['bydate','all'], Help=>'Create one master superflat frame per night [bydate] or from all the images in the list [all] ?'},
   combine => { Value=> 'avsig', Allowed=>['average','median','avsig'], Help=>''},
   scale => { Value=> 'median', Allowed=>['none','median','exposure'], Help=>''},
   sigma=> { Value=> '2', Help=>'Sigma rejection if combine=avsig'},
   niter=> { Value=> '3', Help=>'Number of iterations if combine=avsig'},
# Normalization of the master frame
   normalize=> { Value=> 'yes', Allowed=>['yes','no'], Help=>'Normalize after combination ?'},
   nmode=> { Value=> 'global', Allowed=>['global','single'],
           Help=>'Normalize over the whole mosaic [g] or on single frames [s]'},
# Mosaic
   mkmosaic=>{ Value=> 'yes', Allowed=>['yes','no'], Help=>'Produce a mosaic ? '},
   done => 0,
   CHECK => ''
  },

 fringe => {
 fields => [qw(apply input output  backsize det_thresh  
        group combine scale  sigma niter window mkmosaic)],
  apply => { Value=> 'no', Allowed=>['yes','no'], Help=>'Compute/apply a fringe frame ?'},
# Computation of the master fringe frame
  backsize => {Value=> '300', Help=>'Master fringe: background mesh size'},
#  det_area => { Value=> '5', Help=>'Master fringe: xxx (object masking)'},
  det_thresh => { Value=> '5', Help=>'Master fringe: detection threshold'},
  #smooth => { Value=> '10', Help=>'Master fringe: width in pixels of the smoothing box'},
# Combination of the fringe frames into the master fringe frame
  group  => { Value=> 'bydate', Allowed=>['bydate','all'], Help=>'Create one master fringe frame per night [bydate] or from all the images in the list [all] ?'},
  combine => { Value=> 'avsig', Allowed=>['average','median','avsig'], Help=>''},
  scale => { Value=> 'exposure', Allowed=>['none','median','exposure'], Help=>''},
   sigma=> { Value=> '2', Help=>'Sigma rejection if combine=avsig'},
   niter=> { Value=> '3', Help=>'Number of iterations if combine=avsig'},
  input => { Value=> '', Help=>'Input list of the images used to compute the fringe frames'},
  output => { Value=> 'Pfringe.lst', Help=>'Output list of fringe frames'},
 # Subtraction of the master fringe frame
 #  section => {Value=> '', Help=>'Fringe subtraction: Section where the scaling factor will be computed'},
  window => { Value=> '5', Help=>'Fringe subtraction: width for fitting (pixels)'},
  #iterations => {Value=>1, Help=>'Fringe subtraction: Number of iterations performed'}, 
   mkmosaic=>{ Value=> 'yes', Allowed=>['yes','no'], Help=>'Produce a mosaic ? '},
   done => 0, 
   CHECK => ''
  },

 science => {
 fields => [qw(apply input output weight wout level mkmosaic)],
  apply => { Value=> 'yes', Allowed=>['yes','no'], Help=>'Process science images ?' },
  input => { Value=> '', Help=>'Input list of science images'},
  output => { Value=> 'Pscience.lst', Help=>'Output list of science images'},
  level => { Value=> 'yes', Allowed=>['yes','no'], Help=>'Make the same background level ?'}, 
# Weight maps
  weight => {Value=> 'no', Allowed=>['yes','no'], Help=>'Produce weight maps ?'},
  wout => { Value=> 'wmap.lst', Help=>''},
# Diagnostics
  mkmosaic=>{ Value=> 'yes', Allowed=>['yes','no'], Help=>'Produce a mosaic ? '},
  done => 0,
  plot => { Value=> 'no', Allowed=>['no'], Help=>'To be implemented'},
   CHECK => ''
 },


};


bless $params, $class;


push @_Configs, \$params;


 return $params;

}


sub Entry
{
   my $self=shift;

   if (@_) { @{ $self->{ENTRIES} } = @_ }
   return @{ $self->{ENTRIES} };

}

my $_current = -1;

sub Reset {
my $class=shift;
   $_current = -1;
}


sub each 
{
my $class=shift;
my $nextindex = ++$_current;

if ($nextindex < @_Configs)
		{
			return bless ${$_Configs[$nextindex]}, $class;
		}
		else
		{
			$_current = -1;
			return undef;
		}

}

sub index
{
my $class=shift;
my $index=shift;
$index--;

return undef if ($index <0 || $index > @_Configs);
return bless ${$_Configs[$index]}, $class;

}


sub Count 
{ 
my $count=$#_Configs+1;
return $count
}

sub LoadParams
  {

my $self=shift;

my $file;

if (@_) {$file=shift}
else {$file=$self->{NAME}};


if ( not open (FILE, $file)) { return}

my ($field, $option);
my $flag_f=0;
my $flag_opt=0;


while (<FILE>) {
 chomp($_);
 
  if ($_ =~ /TOSTEP/) {
     my @s=split(/ /, $_); 
     if ($#s==2) {
          $self->{TOSTEP}=$s[1];
	  next
      }
   }   
     
 
   next if ($_ =~ /^#/ || $_ !~ /\w+/);   # Comment or no alphanumeric

   if ($_ =~ /<\w+/) { 
       my $t=$_;
       $t=~ /(<)(\w+)(>)/;
          if ($flag_f==1) {        #START of an OPTION section
	         $option=$2;
		 $flag_opt=1
	         } 
	  else{ $flag_f=1; $field=lc($2)};   #START of a FIELD section
          next
	}  #set FIELD/OPTION begin
   elsif ($_ =~ /\<\/\w+/) {  
          if ($flag_opt==1) {            #END of an OPTION section
	         $option='';
		 $flag_opt=0
		}  
	  else {$field=''; $flag_f=0}    #END of a  FIELD  section
	  next
       }  #set FIELD/OPTION end
   
 #ok, these are the infos to read....  

 
    my @vals =split(" ",$_);

     if ($field eq 'general' && $vals[0] eq 'instrument') {
        $self->Instrument($vals[1]);
     } 
    elsif ($option eq 'section' && $field ne 'general') {
      next if ($#vals !=2);
      my $ccd=$vals[0];
      my @s=split(':',$vals[2]);
    
      $self->Section(step=>$field, ccd=>$vals[0], 
            c1=>$s[0], c2=>$s[1], r1=>$s[2], r2=>$s[3]);
            
      next;
       }
    
    if ($#vals<=1) {
       $option=$vals[0]; 
       my $value='';
       if ($#vals==1) {$value=$vals[1]}

     if (exists $self->{$field}{$option}{Value}) {
        $self->{$field}{$option}{Value} =  $value;
     }	
   }


 } #end while	

 close (FILE);


} #end LoadParams



sub SaveParams
  {
my $self=shift;

my $file;

if (@_) {$file=shift}
else {$file=$self->{NAME}};

if (! open (FILE, ">$file")){
 warn "IConfig::SaveParams - file $file can't be saved!";
 return 1;
}

print FILE "<TOSTEP> ", $self->{TOSTEP}, " </TOSTEP>\n\n";

foreach my $field (@{$self->{steps}})  {

 print FILE "<$field>\n";
   foreach my $option ( @{$self->{$field}{fields}} ) {
      if (exists $self->{$field}{$option}{Value}) {
         print FILE "  ",$option, "  ",$self->{$field}{$option}{Value} ,"\n";
        }
      
       }
       
 if (exists $self->{$field}{section} && $field ne 'general') {
       my $i=0;
	 my @s=$self->Section(step=>$field);
	 my @id=$self->CCDid();
	 print FILE "  <section>\n"; 
	 print FILE "#  seq  id   c1:c2:r1:r2\n";	
	 foreach my $s (@s) {
	   print FILE "     $i   $id[$i]  ",
	      $s->[0],":",$s->[1],":",$s->[2],":",$s->[3],"\n";
           $i++;
	 }              
	 print FILE "  </section>\n"; 		
     }
     
  print FILE "</$field>\n\n";   

last if ($field eq $self->{TOSTEP});
}

close (FILE);


} #end SaveParams

sub getValue
{
 my $self=shift;
 my ($field, $option)=@_;
    return $self->{$field}{$option}{Value};
}


sub mkName
#creates the name of the configuration file in a sequential order
  {
 my $tn="";
 my $root=$main::CONFIG;
 my $ext= $main::CONFEXT;
 my $add='';
 my $nc=0;
 if ($#_Configs==-1) {
     $nc=1;
  }
 else {
   $_= ${$_Configs[-1]}->Name;
   my @s=/(.*[a-z\W]+)(\d+)(.*)(\.)(.*)/;
   if ($#s==4) {       
       $root=$s[0]; $nc=$s[1]+1; $add=$s[2]; $ext=$s[4]}
   else {$nc=$#_Configs+2}
 } 
 $tn=sprintf("%s%02d%s.%s",$root,$nc,$add, $ext);
 return $tn;
  } #end sub mkName

sub Name
#set/return the name of the configuration file
{
  my $self=shift;
 if (@_) { $self->{NAME} = shift}
   return $self->{NAME};
}



sub CheckBox
#Set/return the address of the checkbox
{
  my $self=shift;
  my $step=shift;
  #my ($check, $button, $apply, $info, $del);

  #my $apply=$self->{$step}{apply}{Value};
  if (@_) {$self->{$step}{CHECK}=shift};
  if (exists $self->{$step}{CHECK}) {return $self->{$step}{CHECK}} else {return -1}
  #$info=$self->{$step}{INFO};
}

sub getSteps
#return the steps which should be done
{
  my $self=shift;

  my @what=();

if  ($self->{TOSTEP} ne 'skip') {

 my $last=0;
 
  foreach my $step (@{$self->{steps}})  {
     last if ($last);
      $last=1 if ($step eq $self->{TOSTEP});

     next if (!exists $self->{$step}{done}) ;
     next if ( ($self->{$step}{apply}{Value} eq 'no') );
     push @what, $step
  }
}

  return @what;
}

sub Done
{
  my $self=shift;
  my $step=shift;

  if (@_ ) {$self->{$step}{done}=shift}
  return $self->{$step}{done};
}

sub CCDpos {
my $self=shift;
my %user=@_;

my @pos=@{$self->{CCDpos}};

if (@_) {
if (exists $user{ARRAY}) {
   $self->{CCDpos} = $user{ARRAY};
}
else {

 my $ccd=-1;
 $ccd=($user{ccd}-$self->{MINCCD}) if (exists $user{ccd}); 
    if ($ccd>=0) {    
# insert a position. Note: perl automatically resizes arrays... ;)    
       if (exists $user{POS}) {
           $pos[$user{POS}]=$ccd;
	   $self->{CCDpos} = [@pos];
	   }
	else {   
	  return $pos[$ccd];
	  
            foreach my $i (0..$#pos) {
                if ($pos[$i]==$user{ccd}) {$ccd=$pos[$i]; last};
            }
            return $ccd; 
       }
    } #end if ccd given
    else {   #only the position is given: return the corresponding CCD
       if (exists $user{POS}) { # we want to know the CCD in a given position                      
	    return -1 if ($user{POS} > $#pos);
	    my $ccd=-1;
	    foreach my $i (0..$#pos) {
	        if ($pos[$i]==$user{ccd}) {$ccd=$pos[$i]; last};
	    }
	    return $ccd;            
        }
    }
  } #end else on ARRAY 
 } #end if arguments
 else {return @pos} 
}

sub CCDid {
my $self=shift;
my %user=@_;

my @ids=@{$self->{CCDid}};

if (@_) {
if (exists $user{ARRAY}) {
   $self->{CCDid} = $user{ARRAY};
}
else {

 my $ccd=-1;
 $ccd=$user{ccd} if (exists $user{ccd}); 
    if ($ccd>=0) {
# insert a position. Note: perl automatically resizes arrays... :)    
       if (exists $user{ID}) {
           $ids[$ccd]=$user{ID};
	   $self->{CCDid} = [@ids];
	   }
       else {
         if ($ccd > $#ids) {return -1}
          else {return  $ids[$ccd]}
     }
    } #end if ccd given
    else {
       if (exists $user{ID}) { # we want the sequential number for that index
            my $seq=-1;
            foreach my $i (0..$#ids) {
                if ($ids[$i]==$user{ID}) {$seq=$i+$self->{MINCCD}; last};
            }
            return $seq;
        }
    }
  } #end else on ARRAY 
 } #end if arguments
 else {return @ids} 
  
}


#sub CCDext {

#die "CCDext called in CONFIG.PM ?????????\n";
#my $self=shift;
#my %user=@_;

## this only makes sens for MEF files.....
#return if ($self->{general}{fits_type}{Value} ne 'MEF'); 

#my @exts=@{$self->{CCDext}};

#if (@_) {
#   if (exists $user{ARRAY}) {
#      $self->{CCDext} = $user{ARRAY};
#   }
# else {

# my $ccd=-1;
# $ccd=$user{ccd} if (exists $user{ccd}); 
#    if ($ccd>=0) {
## insert a position. Note: perl automatically resizes arrays... :)    
#       if (exists $user{EXT}) {
#           $exts[$ccd]=$user{EXT};
#	   $self->{CCDext} = [@exts];
#	   }
#       else {
#         if ($ccd > $#exts) {return -1}
#          else {return  $exts[$ccd]}
#     }
#    } #end if ccd given
#  } #end on ARRAY     
#  } #end if arguments
# else {return @exts} 
  
#}



sub Section {
my $self=shift;
my %user=@_;

#at least the step is required !
return if (! exists $user{step} );

my $step=$user{step};

 my $rec=$self->{$step}{section};


if ( exists $user{array}) {
#0. an array is given, so just feed this into 
  my $allsec=$user{array};
    my @sections=();  #empty the array
  foreach my $i (0..$#{$allsec}) {
      my $sec={};
      $sec->{c1}= $allsec->[$i][0];
      $sec->{c2}= $allsec->[$i][1];
      $sec->{r1}= $allsec->[$i][2];
      $sec->{r2}= $allsec->[$i][3];
      push  @sections, $sec;
   }
#ok, now store the array
  $self->{$step}{section}=[@sections];   
}

elsif (exists $user{ccd} || $step eq 'general') {
# 1> the CCD id was given

 my $ccd;

 if ($step eq 'general') {$ccd=0}
 else {$ccd=$user{ccd}}

 if (exists $user{c1} && exists $user{c2} && exists $user{r1} && exists $user{r2})
 {
 
  my @sections=(); 
# 1a> The section was given, so it must be modified  
  my $set=0;
      my $sec={};
      $sec->{c1}= $user{c1};
      $sec->{c2}= $user{c2};
      $sec->{r1}= $user{r1};
      $sec->{r2}= $user{r2};
#if the array contains something, then try to substitute the entry with that ccd
   $rec->[$ccd]=$sec;    
   
 }

#1b> return the section as an array
   my @osec=();

   my $rec=$self->{$step}{section};
   my $xsec=$rec->[$ccd];
   #foreach my $sec (@allsec) {
   #  next if ($sec->{ccd}) != $user{ccd};
     @osec=($xsec->{c1}, $xsec->{c2}, $xsec->{r1}, $xsec->{r2});
   #}

 return @osec;
 }
else {
# 2> return the whole stuff (CCDs and sections)
   my @allsec=();
   
   foreach my $sec (@{$self->{$step}{section}}) {
     my @sec=($sec->{c1}, $sec->{c2}, $sec->{r1}, $sec->{r2});
     push @allsec, [@sec];
   }

 return @allsec;
 } 
 
} #end sub Section


sub getIdx 
  {
my $self=shift;
my $idx=-1;
foreach my $i (0..$#_Configs) {
warn "setidx: index $i, $self - ${$_Configs[$i]}";
  if ($self eq ${$_Configs[$i]}) {$idx=$i; last}
}
warn "setidx returns $idx";
return $idx;
  }

sub Remove
{
my $self=shift;
my $index;
if (@_) {$index=shift} else {$index=$self->getIdx};
my $name=$self->Name;
warn "trying to remove $self (=> $index/$#_Configs)",$name ;
return 0 if ($#_Configs==0 || $index <0 || $index>$#_Configs);
my $w=sprintf("ok, removing %s", $name);
warn $w;
 splice @_Configs, $index, 1;

 return $index;

}

#sub DESTROY {
#   warn "DESTROY called !!!";
#}

# ==========================================================
#      Test functions
# ==========================================================



# ==========================================================
#      List functions
# ==========================================================

sub getList
#Return the name of the input/output list
{
  my $self=shift;
  my $step=shift;
  my $mode;
  if (@_ ) {$mode=shift} else {$mode='input'};
  return $self->{$step}{$mode}{Value};
}


sub checkList
# check the existence of files in a list
{

my $self=shift;

my ($step, $mode)= @_;

my $list=$self->{$step}{$mode}{Value};
my $ok=0;
my $ntot=0;

 $list=~tr/@//d;

 if (-e $list) {
  open (ILIST, "$list");
   while (<ILIST>) {
       my @info=split(/\t/, $_);
       my $file=$info[0]; chomp ($file); 
       $file=~tr/ //d;
       $ntot++; if (-e $file) {$ok++}}
  close (ILIST);
 } else {$ok=-1}

#if ($field eq 'output' && $ok == $INSTR_TAB{$INSTRUMENT}->{NUMCCD}) {$procStatus{$step}{DONE} = 'yes'}


  return ($ok, $ntot);
} #end sub  checkList


sub Instrument {
my $self=shift;
 my $ninstr;


if (@_ ) {
  $ninstr=shift;
  
my $instr;

my $tname=lc($ninstr);

if (! eval "require Instr::$tname") {
  warn "instrument $ninstr not defined !";
  
return -1;
}


eval "\$instr=$tname->new()";

$instr->init_props();    #Set all parameters (properties)

if ($main::DEBUG) {print "#DEBUG Instrument: setting properties for $ninstr\n"}    
         
# these must be done always
      $self->{general}{instrument}{Value}=$ninstr;

       $self->{MINCCD} = $instr->getProps('MINCCD');
       $self->{NCCDX} =  $instr->getProps('NCCDX');
       $self->{NCCDY} =  $instr->getProps('NCCDY');
       $self->{NUMCCD} = $instr->getProps('NUMCCD');
       $self->{general}{fits_type}{Value}=$instr->getProps('FITS_TYPE');

	if ($instr->getProps('overscan','order') ne '') {
          $self->{overscan}{order}{Value} = $instr->getProps('overscan','order')
	}  
        if ($instr->getProps('overscan','method') ne '') {
          $self->{overscan}{method}{Value} = $instr->getProps('overscan','method')
	} 

       	if ($instr->getProps('fringe','group') ne '') {
          $self->{fringe}{group}{Value} = $instr->getProps('fringe','group')
	}  
	if ($instr->getProps('sflat','group') ne '') {
          $self->{sflat}{group}{Value} = $instr->getProps('sflat','group')
	} 


       my @secs;
       
       @secs=$instr->Section(step=>'data');
       $self->Section(step=>'general', c1=>$secs[0], c2=>$secs[1], r1=>$secs[2], r2=>$secs[3]);
              
       @secs=$instr->Section(step=>'trim');
       $self->Section(step=>'trim', array=>\@secs);
       
       @secs=$instr->Section(step=>'overscan');
       $self->Section(step=>'overscan', array=>\@secs);
     
       my $ids=$instr->getProps('CCDid');
       $self->CCDid(ARRAY=>$ids);
       
       my @pos=$instr->getProps('CCDpos');
       $self->CCDpos(ARRAY=>@pos);
       
       my $fits=$instr->getProps('FITS_TYPE');
  
   } #else if input
        

return $self->{general}{instrument}{Value};

}

}

1;
