#============= Last modified =============================================
#       May 9 2001 (added: object detection with SEx in fringe correction)
#=========================================================================

package PDLfuncs;


use PDL;
use Astro::FITS::CFITSIO qw( :longnames :constants);
use PGPLOT;


use Mylib::Pfit;
use Mylib::Prered;
use Mylib::Robust;

use PDL::Image2D;
use  PDL::Graphics::PGPLOT;

use File::Basename;
use Cwd;
use Common::misc;
use Common::fitsio;

use Instr::Header;

use Tie::IxHash;

use Shell qw(df);
use Shell qw(top);

use subs 'rfits', 'wfits';
use Sex::RunSEx;
use WW::RunWW;

use strict;

# This is needed in order to correctly send infos to the GUI
if (defined  $main::interface) {
    if ($main::interface eq 'gui') {require Tk}
}

sub domosaic {
  #make a mosaic, actually is a simplified 'combine' 
  # format: domosaic($list, $params, {LEVEL=>'yes', UPDATE=>'no', MKMOS=>'yes', ...})
  
  my $list=shift;
  my $params=shift;
  my $opt; $opt = pop @_ if ref($_[$#_]) eq 'HASH';
  
  my @bin=();
  my @pos=();
  my $nim=0; 
  return if ($main::redmode) eq 'listonly';


 my @ifiles=();
  
  open (IFILE, $list);
  while (<IFILE>) {
    my @info=split(/\t/, $_);
    my $file=$info[0]; chomp($file);
    $file=~tr/ //d;
    push @ifiles, $file if (-e $file);
  }
  close (IFILE);
  
  
  #1) Should I build the mosaic ?
  my $mkmos=0;
  if ($opt->{MKMOS} eq 'yes') {
    $mkmos=1;
    @bin=($params->{mosaic}{binx}{Value},  $params->{mosaic}{biny}{Value});    
    #@pos=$params->CCDpos();
  }
  
  
  #2) Should I bring the background to the same level ?
  my $bckcalc=0;
  my $bckapply=0;
  my ($badvalue, $sigma, $niter); #used to compute the background
  
  if ($opt->{LEVEL} =~ /^y/i) {
    $bckcalc=1; 
    $badvalue=$params->{mask}{badvalue}{Value};
    $sigma=$params->{mosaic}{sigma}{Value};
    $niter=$params->{mosaic}{iterations}{Value};
    
    #2a) Should I write the files ?
    if ($opt->{UPDATE} =~ /^y/i) {
      $bckapply=1; 
    }
  }
  
  
  my %mh;
  tie (%mh, 'Tie::IxHash');  
  
  my $HEAD=Header->new();

my $mproc=0;
  
my $nfiles=$#ifiles;

 FILES: while (@ifiles) {
my $pc=int(100*($nfiles-$#ifiles)/$nfiles+0.5);
    
    my $ffile=$ifiles[0];

    
    #build the name of the mosaic  
    my $name;
    if ($opt->{MKMOS} eq 'yes') {  
       $mkmos=1;
      %mh=&fitsio::rhead($ffile);   

      @pos=();
      $nim=-1;
      
      $name=$HEAD->buildName({NAME=>$ffile, FLAGS=>'mos'}); #keep the same name, but change flags and CCD

      if ($name eq '') {return 1} #the name could not be built !

      if (-e $name) {
	if ($main::REDO eq 'yes') {system ("rm $name")}
	else {$mkmos=0}
       }

    } 

    #find the root in the filename     
    my @r=split(/#/, $ffile, 2);
		if ($#r <1) {
		  $_=$ffile; 
		  @r=/(.*[a-z\W]+)(\d+)(.*)(\.fits)$/;
		  #print "$ffile / $r[0]\n";
		  if ($#r < 2) {shift @ifiles ; next}
		}
		my $root=$r[0];
		
		#extract the filenames with the same root 
		my @list=grep(/$root/, @ifiles);
		if ($#list <=0) {shift @ifiles; next FILES};
		@ifiles=grep(!/$root/, @ifiles);      
		
		#Ok, loop on the files with the same root      
		
		my ($stack, $bavg, $brms);
		my $k=-1;
	      LIST:   for my $image (@list) {     
		 
		  $k++;
		  if ($main::DEBUG) {print "#DEBUG MOSAIC: reading $image\n"}
		  print "counter $pc $image\n" if ($main::interface eq 'gui');	
		  
		  if ($main::redmode ne 'RUN') {$mproc++; next LIST};
		  
		  next LIST if (!-e $image);
		  $mproc++;

	  

		  my $input=null;

		  my %lh=&fitsio::rhead($image);   
      		  my $lchip=$lh{IMAGEID};
                  $pos[++$nim] = $params->CCDpos(ccd=>$lchip);

		  if ($bckcalc) { #should be the background computed ?
		    
		    if ($k==0) {
		      $bavg=zeroes(float, $#list+1); 
		      $brms=zeroes(float, $#list+1); 
		    }
		    my ($avim, $sim);
		    $avim=0; $sim=0;

		   if (exists $lh{BACKSTAT}) {
			my $msg="$image: BACKSTAT keyword found";
			&main::mkLog($msg,"W",'level');
			my @s=split(/ /, $lh{BACKSTAT}, 4);
			$avim=$s[1]; $sim=$s[2];
		      }
		    
		    if ($avim==0 || $mkmos==1) {
#ok, read the image
		            my $ret;
		           ($input, $ret)=&fitsio::rfits($image);
			   next LIST if ($ret);
			 }

		    if ($avim==0) {
		        my $im1d=$input->clump(-1);
		        ($avim, $sim) = avsigclip($im1d, $sigma, $niter, $badvalue);
		      }
		    set $bavg, $k, $avim; #store the values for later
		    set $brms, $k, $sim;  	  	  
		  }
		  
		  if ($mkmos) {
		    if ($input->isempty) {
                           my $ret;
		           ($input, $ret)=&fitsio::rfits($image);
			   next LIST if ($ret);
                      }
		    my $out=&Mylib::Prered::blkavg($input,{BIN=>\@bin});
		    if ($k==0) {
		      my ($nx, $ny)=$out->dims;
		      $stack=zeroes(float, $nx, $ny, $#list+1);
		    }	     
		    my $tmp=$stack->slice(":,:,($k)");
		    $tmp.=$out;		    
		  }
		  
		}		#end LIST
		
                next FILES if ($main::redmode ne 'RUN');

		if ($bckcalc) {	#bring all images to the same background level
		  my $nzidx=which($bavg != 0); #just in case some image is missing....
		  my $globavg=average($bavg->index($nzidx));
		  if ($main::DEBUG) {print "#DEBUG individual averages: $bavg / global average: $globavg\n";}
		  
		  my $g_bavg=$bavg-$globavg;


		  if ($mkmos) {	#Correct the mosaic
		    my $tmp = $stack->mv(2,0);		    
		    $tmp -= $g_bavg;
		  } 
		  
		  if ($bckapply) { #Rewrite images and update headers
		    $k=-1;
		  UPLIST:   for my $image (@list) {   		      
		      $k++;
		      my %fh;
		      tie (%fh, 'Tie::IxHash'); 
		      %fh=&fitsio::rhead($image);
		   		     
		      next UPLIST  if (exists $fh{BACKSTAT});

		      my ($input, $ret)=&fitsio::rfits($image);
		      next UPLIST if ($ret);
		      
		      my $comm= sprintf "Mean %.3f RMS %.3f",$bavg->at($k), $brms->at($k);
		      chomp($comm);
		      
		      if ($main::DEBUG) {print "#DEBUG: $k ", $bavg->at($k), "  " ,$brms->at($k),"\n"}
		      
		      $fh{BACKSTAT} = $comm;
		      $fh{COMMENTS}{BACKSTAT} = "The average and variance of the background";
		      
		      $input-=$g_bavg->at($k);
		      $comm= sprintf "%.3f",$globavg;
		      chomp($comm);
		      $fh{BACKADD} = $comm;
		      $fh{COMMENTS}{BACKADD} = "The average background in the mosaic";
		      
		      if ($main::DEBUG) {
			print "#DEBUG setting background level for image $image\n";	
			print "#DEBUG $comm\n";
		      }	 	    
                    $main::_wfits=$image ;  
		      &fitsio::wfits($input, $image, \%fh); 
                   $main::_wfits=''; 
		      
		    }		#end of loop on UPLIST 	  	  	  
		  }		#end if bckapply
		  
		}		#end if bckcalc
		
		if ($mkmos) {	#OK, make the mosaic !!!!
		  my $panels=[$params->{NCCDX},$params->{NCCDY}] ;
		  my $mos;
		  
		  $mos=&Mylib::Prered::mosaic($stack,{PANELS=>$panels, POS=>\@pos}) ;
		  
		  if ($main::DEBUG) {print "#DEBUG writing mosaic $name\n"}

                $main::_wfits=$name ;  
		  &fitsio::wfits($mos, $name, \%mh);
                $main::_wfits=''; 
		}	  
		
	      }			#end FILES

return $mproc;    

  }				#end sub domosaic
  
  sub combine {
    
    my ($inlist, $outlist, $step)= @_;
    
    my $nerr=0;
    my $wfiles=0;        #the files written on disk
    my $procfiles=0;     #the files in the list
    my $dfiles=0;        # the files already existent

    my @ifiles=();    
    my @ofiles=();
    
    my $bydate='no';
    $bydate='yes' if (
        ($step eq 'fringe' &&  $main::params->{fringe}{group}{Value} eq 'bydate') ||
        ($step eq 'sflat' &&  $main::params->{sflat}{group}{Value} eq 'bydate')
      );    


    my $bproc=0;

    # ---- Where the combined files will be stored
    my $todir=$main::STOREDIR;
    
    $inlist =~ tr/@//d;
    $outlist =~ tr/@//d;
    
    my $SCALE = $main::params->{$step}{scale}{Value};
    my $COMBINE = $main::params->{$step}{combine}{Value};
    my $SIGMA= $main::params->{$step}{sigma}{Value};
    my $NITER= $main::params->{$step}{niter}{Value};
    my $BADVAL=$main::params->{mask}{badvalue}{Value};
    
    my $l="Combining frames from $inlist ";
    if (! -e $inlist) { $l=$l."  (NOT FOUND !!!)"};
    $l=$l." to <out>$outlist</out>\n";
    
    $l=$l."    WITH scale: $SCALE, combine: $COMBINE";
    
    &main::mkLog($l);
    
    
    if ($main::redmode ne "listonly") {
      
      if (-e $outlist) {system ("rm $outlist")}
      
      
      open (IFILE, $inlist);
      while (<IFILE>) {	 
	 chomp($_);
	 push @ifiles, $_;     #this actually includes file name AND optional infos
       }
      close (IFILE);
      
      if ($main::interface eq 'gui') {
	print "counter 0 none \n";
      }
      
      my $last=0;
      
      my $name='';
      
      my $HEAD=Header->new();
      
      my @xfiles=@ifiles;	#this array will be progressively emptied...
      
      my @pfiles=();    #flags telling if a file was already processed

# == initialize the pfiles array
      foreach my $i (0..$#ifiles) {$pfiles[$i]=0}

    FILES: foreach my $i (0..$#ifiles) {

     next if ($pfiles[$i]!=0);
	
	my $done=0;		#this will be set to 1 if no combination is required
	        
        my @info=split(/\t/, $ifiles[$i]);         
	my $ffile=$info[0]; chomp ($ffile);
        $ffile =~ tr/ //d;
	#chomp($ffile);

warn "---> Searching fellows for $ffile" if ($main::DEBUG);	

print "counter 0 $ffile\n"; #shows the reference file for which fellows are searched

	@xfiles=grep (!/$ffile/, @xfiles);
        $pfiles[$i]=1;	

	if ($ffile !~ /\w+/) {last}
	
	if (! -e $ffile && $main::redmode ne "mklists") {
	  my $msg="Could not open file $ffile";
	  &main::mkLog($msg,"E",$step);
	  $nerr++;
          $pfiles[$i]=-1;
	  next FILES;		#skip to next file in the list
	}
	
	my (%fh,$fchip, $ldate) ;
 
       if ($main::redmode ne "mklists") {       
	  tie (%fh, 'Tie::IxHash'); #this header will be written later, so we need to preserve order
	  %fh=&fitsio::rhead($ffile);	

	$fchip=$fh{IMAGEID};	#READ THE NUMBER OF THE CCD (may be different for other telescopes ?!?)
	$ldate=$fh{LOCDATE}  if ($bydate eq 'yes');
       
       } else {  
#the files do not exist, so we must obtain the infos from the list
	  my @s=split(/\./, $ffile);

           $fchip=$s[-2];
	   $fchip=~tr/#//d;
           $ldate=$info[1];	  
	   $fh{DETECTOR}=$s[1];
	   $fh{FILTER}=$s[2];
	   $fh{IMAGEID}=$fchip;
	   $fh{LOCDATE}=$ldate;

       }	


	$name=$HEAD->buildName(\%fh, {STEP=>$step, OUTDIR=>$todir, DATE=>$bydate});



	if ($name eq '') {
	  my $msg="Could not build name from image $ffile ?!?";
	  &main::mkLog($msg,"E",$step);
	  return;
	}			#the name could not be built

	if ($main::DEBUG) {print "#DEBUG $name: redo = ", $main::REDO , "\n";}
	if (-e $name) {
	  if ($main::REDO eq 'yes' && $main::redmode eq 'RUN') {system ("rm $name")} else {
	    my $msg="Image $name exists - skipping";
	    &main::mkLog($msg,"W",$step);
	    $done=1;
            $dfiles++;
	  }
	}
	
	#if (! $done) {
	# Initialize the list containing all the files to be combined;
	#  add the current (first) image only if not saturated
	my @list=();
	
        if ($main::params->{select}{xsatur}{Value} eq 'yes') {
	  if (exists $fh{SATUR} && $fh{SATUR}==1) {	
	    my $msg="Image $ffile is saturated, skipping !";
	    &main::mkLog($msg,"W",$step);
	  }	 else {push @list, $ffile}
	}
	
	# Ok, now feed the list with other images
	
      MATCH:    foreach my $j ($i+1..$#ifiles) {

	next if ($pfiles[$j]!=0) ; #this file was already processed  

	  my @minfo=split(/\t/, $ifiles[$j]);         	 
	  my $nfile=$minfo[0];
	  $nfile =~ tr/ //d;
	  #chomp($nfile);
	  
	  if ($nfile !~ /\w+/) {last}

	my (%nh, $nchip, $nldate);

	if ($main::redmode ne "mklists") {
	  if (not -e $nfile) {
	    my $msg="Could not open file $nfile";
	    &main::mkLog($msg,"E",$step);
	    $nerr++;
	    $pfiles[$j]=-1;
	    next MATCH;
          }
	  
	  %nh=&fitsio::rhead($nfile); #no need to preserve order !
	  $nchip=$nh{IMAGEID}; #READ THE NUMBER OF THE CCD 
	  $nldate=$nh{LOCDATE};
	} else {
	   my @s=split(/\./, $nfile);
	   $nchip=$s[-2];
	   $nchip=~tr/#//d;
	   $nldate=$minfo[1];
	}

	  #if the DATE option is ON, then check if the date is the same
	  if ($bydate eq 'yes') {	  
#warn "comparison $nldate vs $ldate";	   
	    next MATCH if ($nldate ne $ldate);
	  }
	  
	  #A frame with same CCD was found
	  if ( ($nchip == $fchip)) {
	    #check for saturation and update lists 
           if ($main::params->{select}{xsatur}{Value} eq 'yes') {
	    if (exists $nh{SATUR} && $nh{SATUR}==1) {	
	      my $msg="Image $nfile is saturated, skipping !";
	      &main::mkLog($msg,"W",$step);
	      $pfiles[$j]=-1;
	      next MATCH;
	    }
           }	

warn "$nfile is a fellow !" if ($main::DEBUG);

	    push(@list,$nfile);	#the array @list contains the filenames matching that CCD
	   
	    @xfiles=grep (!/$nfile/, @xfiles);
	    $pfiles[$j]=1;
	  }			#end if
	}
	
	if ($#xfiles==-1) {$last=1}	  
	
	
	if ($main::DEBUG) {print "#DEBUG list for $fchip contains  $#list files\n"}
	
	if ($#list==-1) {
	  my $msg="NO files to combine for chip $fchip!";
	  &main::mkLog($msg,"E",$step);
	  next FILES;
	}
	
	
	# (a) there are more files than CCDs, but no combination is required:
	#     simply copy the list of files to the final one and return
	
	if ($#list >0 && $COMBINE =~ /no/i) {	
	  system ("cp $inlist $outlist");
	  return 0;
	}
	
	# (b) there is only one file per CCD: change the name and compute the average,
	# 
	
	if ($#list == 0)  {
	  if ($COMBINE !~ /no/i) {
	    my $msg="Only 1 frame to combine";
	    &main::mkLog($msg,"W",$step);
	  }

	  if ($done==0  && $main::redmode ne "mklists") {
	  my %fh;
	  tie (%fh, 'Tie::IxHash'); #this header will be written later, so we need to preserve order
	  %fh=&fitsio::rhead($ffile);
	  my ($image, $ret) = &fitsio::rfits($list[0]);
	  my ($mean, $rms,$skew, $kurt, $min, $max)=moments($image->clump(-1));
	  my $comm=sprintf "Mean %.3f RMS %.3f Skew %.3f Kurt %.3f",$mean, $rms,$skew, $kurt;
	  chomp ($comm);
	  $fh{'STAT'}=$comm;
	  if ($mean > $main::params->{select}{highrej}{Value} ) {
	    my $msg=sprintf("Image %s is saturated (avg >= %d)",$list[0],$main::params->{select}{highrej}{Value});
	    &main::mkLog($msg,"W",$step);
	    $fh{SATUR} = 1;
	  } #end if only 1 file
          
#print "---> $name";

          $main::_wfits=$name;  
	     $ret=&fitsio::wfits($image, $name, \%fh);
          $main::_wfits='';
	  $wfiles++;
#print "  ... $ret\n ok";

	  if ($ret ne "") {&main::mkLog($ret,"E",$step);$nerr++;return $nerr;}
	  else {$done=1};
	}

	  	  
	}


	if ($done==0 && $main::redmode ne "mklists") {
	  
	  # (c) ok, go ahead !
	  
          #Create a stack of images
	  if ($main::DEBUG) {print "#DEBUG reading image $list[0]\n"}
	  my ($image, $ret) = &fitsio::rfits($list[0]);	#read the first image of the list
          if ($image->isempty) {my $msg="$list[0] - $ret";
				&main::mkLog($msg,"E",$step);
				$nerr++;
				return $nerr;
			      }	
	  
	  #Now check how much memory it is taking
	  # Each image of dimension ($nx,$ny) will be splitted in ($blocks) blocks of dimension ($deltax,$ny)
	  
	  my $sp=$image->info("%M");
	  my @arr=split(/mb/i,$sp,2);
	  $sp=($#list+1)*$arr[0];    #total space taken in memory by all images
	  
	  my ($nx, $ny) = $image->dims;	#Image dimension
	  
	  my $blocks=1;		#The number of blocks in which the image should be divided
	  
	  #my $maxdim=&checkmem;    #maximum space in Mb: take 1/4 of the total physical memory to avoid swap problems
	  
          my $maxdim=$main::MAXMEM;
	  
	  if ($main::DEBUG) {print "#DEBUG MAXDIM: $maxdim\n";}	  	
	  my $ncim;
	  
	  if ($sp >= $maxdim) { 	
	    $blocks=$sp/$maxdim; #e.g. 60 images of 32Mb => each may take 1Gb/60=16.7Mb, that is should be splitted in 1.92 parts
	    $blocks=int($blocks+1); #take the next integer
	    $ncim= zeroes($image->type,$nx,$ny);
	  }
	  
	  my $deltax=int($nx/$blocks+0.5); #This will be the width of each block
	  
	  if ($blocks > 1 && $main::DEBUG) {print "#DEBUG WARNING: storing all images would take $sp Mb, maximum = $maxdim Mb: they will be splitted in $blocks blocks, nx = $deltax/$nx, ny=$ny\n";}
	  
	  my $cim;		#this will be a huge array, make sure that it is deallocated	
	  my $scalf=zeroes($image->type,$#list+1);    # This array will contain the scaling factors
	  
	BLOCK:  for my $nblock (1..$blocks) {	
	    
	    my @xsec=( 1+($nblock-1)*$deltax, $nblock*$deltax);	# The x-section to be read
	    if ($xsec[1] > $nx) {$xsec[1] = $nx} # the last block is usually smaller than the others
	    my $secnx=$xsec[1]-$xsec[0]+1; # Actual size of each block in x
	    my $sec="[".$xsec[0].":". $xsec[1].",*]"; # OK, this is the section to be read
	    if ($main::DEBUG) {print "#DEBUG Block $nblock => $sec\n"}
	    
	    my $stack= zeroes($image->type,$secnx,$ny,$#list+1);   #build a stack for all blocks          	
	    
	    # now fill $stack with the image blocks in the list	  	
	  LIST:    for my $k (0..$#list) {
      
          
         my @s=split('/',  $list[$k]);

	   
	    $bproc++;  #the number of blocks processed
	    my $pc=int(100*$bproc/(($#ifiles+1)*$blocks)+0.5);

#warn "$bproc $#ifiles $blocks $pc";

	    if ($main::interface eq 'gui') {             
	      print "counter $pc $s[-1]($nblock#$blocks)\n";	     
	    } else {
	      printf "Processed files: %.0f%s\n",$pc,'%'
	    }  


	      my $filesec=$list[$k].$sec; # Filename+section to be read	
	      
              if ($k > 0 || $blocks > 1) {
		($image, $ret)=&fitsio::rfits($filesec,$secnx,$ny); # Read the section, we must give also its size
		if ($main::DEBUG) {print "#DEBUG Image $filesec READ ! \n"};
		if ($image->isempty) {my $msg="$list[$k] - $ret";
				      &main::mkLog($msg,"E",$step);
				      $nerr++;
				      next LIST;
				    }	
		
              }
	      
	      if ($main::DEBUG) {print "#DEBUG Filling block $nblock of image #$k/$#list\n"}	     		 	
	      my $tmp;		
	      ($tmp=$stack->slice(":,:,$k")) .= $image;	#fill with $image the k-th position of $stack
	      
	      #Fill the array containing scaling factors
	      if ($SCALE eq "exposure" || $SCALE eq "median") {
		# in this cases we only need to do it once for each image
		if ($nblock==1) {
		  if ($main::DEBUG) {print "#DEBUG Computing scaling factor ($SCALE)\n"}	     		
		  if ($SCALE eq "exposure") {	
		    my %h=&fitsio::rhead($list[$k]);
		    #set ($scalf,$k,$h{EXPTIME});
		    PDL::Core::set_c ($scalf, [$k], $h{EXPTIME}); #this should be a bit faster
		  }		#end of exptime
		  elsif  ($SCALE eq "median") {
		    my $m;
		    if ($blocks !=1 ) {	# Unless the whole image was already read....
		      my $ret;
		      # ... we must read the whole image.....
		      ($image, $ret)=&fitsio::rfits($list[$k]);
		      if ($image->isempty) {my $msg="$list[$k] - $ret";
					    &main::mkLog($msg,"E",$step);
					    $nerr++;
					    next LIST;
					  }	
		    }
		    $m=medover($image->clump(2));
		    if ($main::DEBUG) {print "#DEBUG scal. factor = $m \n"};			
		    #set ($scalf,$k,$m);
		    PDL::Core::set_c ($scalf, [$k], $m); #this should be a bit faster
		    
		  }		#end of median
		}		#end of check on nblock
	      }			#end if Exptype/Median
	      elsif  ( $SCALE eq "average") { #in this case we can do it for each block
		my $a=($scalf->at($k)) + ( sumover($image->clump(2))/($nx*$ny) );
		#set ($scalf,$k,$a);
		PDL::Core::set_c ($scalf, [$k], $a); #this should be a bit faster
		
	      }
	      elsif  ( $SCALE eq "none") {$scalf .=1}
	      else {my $msg="Unknown scaling type: $SCALE";
		    &main::mkLog($msg,"E",$step);
		    $nerr++;
		    return ($nerr, 0)}
              #end of scaling factors 		   	
	      
	    }			#end of loop LIST
	    
	    if ($main::DEBUG) {print "#DEBUG Scaling factors COMPUTED \n"}
	    
	    #check how frames should be SCALED
	    #use threading to compute Ij=Ij / sf(Ij) * avg(sf(Ij)j=1,N)	, e.g.: sf=med
	    if ($SCALE ne "none") {
	      my $av_sfac= avg($scalf);	# Average of scaling factors
	      my $n_sfac= $scalf/$av_sfac ; # Normalized scaling factors
	      my $tmp= $stack->mv(2,0) ; # $tmp is a virtual PDL pointing to $stack
	      $tmp /= $n_sfac;	# FAST INPLACE backward assigment to $stack
	      if ($SCALE eq "exposure" && $nblock==1) {$fh{EXPTIME}=$av_sfac};
	    }		
	    
	    if ($main::DEBUG) {print "#DEBUG combining with: $COMBINE\n";}
	    #check how frames should be COMBINED over 3rd dimension (must be moved to 1st to use threading)             	
	    if ($COMBINE eq "median") {	#"Simple" median
	      if ($main::DEBUG) {print "#DEBUG Computing median \n"};
	      my $xstack=
                medover($stack->mv(2,0),($cim=null));
	    }
	    elsif ($COMBINE eq "average") { #"Simple" average
	      if ($main::DEBUG) {print "#DEBUG Averaging \n"};	
	      average($stack->mv(2,0),($cim=null));
	    }
	    elsif ($COMBINE eq "avsig") { #Average with sigma clipping		
	      if ($main::DEBUG) {print "#DEBUG Average with Sigma clipping ($SIGMA) \n"}
	      my $rms;
	      ($cim, $rms)=avsigclip($stack->mv(2,0), $SIGMA, $NITER, $BADVAL);
	    }
	    else {my $msg="Unknown combination type: $COMBINE";
		  &main::mkLog($msg,"E",$step);
		  $nerr++;
		  return ($nerr,0);}
	    
	    if ($main::DEBUG) {print "#DEBUG Block STACKED \n"};		
	    
	    if ($blocks > 1) {
	      #Insert the combined block ($cim) into the right section of the final image ($ncim)
	      my $xa=$xsec[0]-1;my $xb=$xsec[1]-1;
	      if ($main::DEBUG) {print "#DEBUG Image section inserted into $xa $xb\n"}
	      
	      #print "Image section inserted into $xa $xb \n";	     	     	
	      my $tmp;
	      ($tmp = $ncim->slice("$xa:$xb,:")) .= $cim;
	    }
		    
	  }			#end of loop BLOCK
	  
	  if ($blocks > 1) {$cim=$ncim}	#Rename the whole image ($ncim) back to $cim
	  
	  #Update the header of the output image and write it

	   #cleanComb(\$fh); #remove keywords from single-image
	  
          my $nframes = $#list+1;
	  my $comm="$nframes frames, scaled by $SCALE, combined by $COMBINE";        	
	  $fh{COMBINE}=$comm;
	  
	  if ($cim->info("%T") =~ /double/i) {
	    my $msg="Output image is double, converted to float";
	    &main::mkLog($msg,"W",$step); $cim=float($cim)
	  }
	  
	  
	  
	  #check that there is enough space on disk !
	  
	  my $df=&checkdf($cim->info("%M"), $name);
	  if ($df ne "") {&main::mkLog($df,"E",$step);$nerr++;return ($nerr, $wfiles);}
	  
	  
	  
	  #---- write on header the images from which the stack was done
	  for my $j (0..$#list) {
	    my ($root, $p) = fileparse($list[$j], '');
	    my $key=sprintf("F_IN%04d",$j);
	    $fh{$key} = $root;	#write the root
	    #extract from the headers of the input files any useful infos
	    my %h=&fitsio::rhead($list[$j]);
	    next if (!exists $h{STAT});
	    #now copy the statistics for each file
	    $key=sprintf("STAT%04d",$j);
	    my $s=$h{STAT};	
	    $fh{$key}=$s;
	  }
	  
	  
	  #OK, write the image !!!!!
	  
	  #if ($main::SAVETYPE eq 'short') {$stype=16}
	  #elsif ($main::SAVETYPE eq 'long') {$stype=32}
	  #elsif ($main::SAVETYPE eq 'double') {$stype=-64}
	  
	  my ($mean, $rms,$skew, $kurt, $min, $max)=moments($cim->clump(-1));
	  $comm=sprintf "Mean %.3f RMS %.3f Skew %.3f Kurt %.3f",$mean, $rms,$skew, $kurt;
	  chomp ($comm);
	  $fh{'STAT'}=$comm;
	  if ($mean > $main::params->{select}{highrej}{Value} ) {
	    my $msg=sprintf("Image %s is saturated (avg >= %d)",$name,$main::params->{select}{highrej}{Value});
	    &main::mkLog($msg,"W",$step);
	    $fh{SATUR} = 1;
	  }

          my $today=&main::mkdate();
          $fh{PROCDATE} = sprintf "%s", $today; 
          $fh{COMMENTS}{PROCDATE} = 'The date when the file was written';         
	  $fh{OUTNAME}=$name;
	  $fh{COMMENTS}{OUTNAME}='The name of the file after preprocessing';
 
          $main::_wfits=$name;  
	     $ret=&fitsio::wfits($cim, $name, \%fh);
          $main::_wfits='';
          $wfiles++;
	  
	  if ($ret) {
	    my $msg=sprintf("Could not write image %s",$name);
	    &main::mkLog($msg,"W",$step);
            $name='';
	  }
	  
	  if ($main::DEBUG) {print "#DEBUG Image $name written ($ret) !!!\n"}
	  
	}			#end if not already done
	
     # } #end of BIG if $name does NOT exist
	if ($name) {
	  print "#DEBUG adding $name to $outlist\n" if ($main::DEBUG);
	  push @ofiles, $name;
	  open (OLIST, ">>$outlist");
	  print OLIST "$name\n";
	  close (OLIST);
	  $procfiles++;
	};
	
	
	#check if the end of a CCD set was reached
	
	
	if ($last == 1) {last}
	
      }				#end loop on FILES
      
      
#      if ($#ofiles+1 != $main::params->{NUMCCD}) {
#	my $msg=sprintf("The output list contains %d images, but they should be %d",
#			$#ofiles+1,$main::params->{NUMCCD});
#	&main::mkLog($msg,"W",$step);
#      }


# ======  Keep track of the list 
  &main::SumList($outlist, 'combine', $step, $procfiles, $dfiles);
 
            
      #Normalize if we are combining flats
      
      if ($wfiles>0 && $step =~ /flat/ && $main::params->{$step}{normalize}{Value} eq 'yes') {
	NormFlats (LIST=>$outlist, FLAT=>$step, DONORM=>'yes');
      } 
 
      return ($nerr, $wfiles);
      
    }				#end if ! listonly
    
    
    
  }				#end sub combine
  
  
  
  sub ccdproc {
    
    my ($inlist, $outlist, $step) = @_;
    
    if ($main::DEBUG) {print "#DEBUG STEP: $step\n"}
    
    #my $step="CCDproc";
    
    $inlist =~ tr/@//d;
    $outlist =~ tr/@//d;
    
    my @ifiles=();
    my @ofiles=();
    my $todir='';


    if ($inlist eq '') {
      my $msg="No input list is given !";
      &main::mkLog($msg,"E",$step);
      return 1;
    }
    elsif (! -e $inlist) {
      my $msg="Could not open list $inlist ";
      &main::mkLog($msg,"E",$step);
      return 1;
    } else {   
      if ($main::DEBUG) {print "#DEBUG reading list $inlist\n"}                
      open (IFILE, $inlist);
      while (<IFILE>) {
	chomp $_;
	push @ifiles, $_ if (-e $_);
      }
      close (IFILE);
      if ($#ifiles==-1) {
        my $msg="The list $inlist contains no existing image";
	&main::mkLog($msg,"E",$step);
	return 1;   
      }
    } 
    
            
    my $instrument=$main::params->{general}{instrument}{Value};
    my $fits_type=$main::params->{general}{fits_type}{Value};
    
    #===========================================================
    #  LOAD THE PROPER INSTRUMENT CLASS
    # ----------------------------------
    my $ins=lc($instrument);
    if (! eval "require Instr::$ins") {
      my $msg="The instrument $instrument is not supported !!\n"; 
      &main::mkLog($msg,"E",$step);
      return;
    }
    
    my $ICLASS;
    if (!eval "\$ICLASS=$ins->new()") {
      my $msg="Something is wrong in the package Instr::$ins !!\n";
      &main::mkLog($msg,"E",$step);
      return;
    };
    
    $ICLASS->init_head();
    #===========================================================
    
       
   if ($main::DEBUG) {print "#DEBUG Building filenames, format: $fits_type\n"}
      
   #Create a list of root file name, if the format is not MEF
   my @roots=();
   my ($minccd, $nccd);
   $minccd=$nccd=0;
   
   if ($fits_type eq 'SIMPLE') {
     
     #------------------------------------------------------------------------
     # SIMPLE FITS files: the list is assumed to be e.g.:
     #        nnnnxx[FF].fits
     #  where nnnn is the ROOT (e.g. /u2/data/toto),
     #        xx is the CCD id number
     #        [FF] are OPTIONAL flags (=everything between digits and .fits)
     #
     #------------------------------------------------------------------------
     
     my @xlist=@ifiles;
     
     $minccd=$nccd=0;
     
     while (@xlist) {
       $_ = $xlist[0];
       
       # separate into root, # of CCD, flags (if present), extensions
       my @string=/(.*[a-z\W]+)(\d+)(.*)(\.fits)$/;
       
       if ($#string < 2) {shift @xlist ; next}
       
       if ($minccd==0) {$minccd=$string[1]};
       my @xs=split(/\//, $string[0]);
       if ($nccd==0) {$nccd=grep(/$xs[$#xs]/, @xlist)}
       if ($main::DEBUG) {print "#DEBUG    $string[0] added to list\n";}
       push (@roots, $string[0]);
       my $odim=$#xlist;
       @xlist=grep(!/$xs[$#xs]/, @xlist);
       if ($odim == $#xlist) {
	 my $msg="Something is wrong in the input list $inlist";
	 &main::mkLog($msg,"E",$step);
	 last;
       }
     }
   } elsif ($fits_type eq 'MEF') {
     #In this case the list files already contain what we need
     foreach (@ifiles) {
       chomp($_);
       my @s=split(/\./,$_);	#find extension
       next if ($#s==-1);
       my $ext=".$s[$#s]";
       my ($base, $p, $s) = fileparse($_, $ext);
       push (@roots, $p.$base);
     }
     $minccd=$main::params->{MINCCD};
     $nccd=$main::params->{NUMCCD};
   }
   
   if ($#roots ==-1) {
     my $msg="No file found in the list";
     &main::mkLog($msg,"E",$step);
     return 1;
   }
   #my $nsteps=0;
   
   
   my $l="Processing frames from <in>$inlist</in>";
     $l=$l." (".($nccd*($#roots+1))." files)";
   
   if ($step eq 'science' ) {
     #This is the last step in case of science frame (no combination performed)
     $todir=$main::STOREDIR;
     $l=$l." to <out>$outlist</out>";
   }
   else {
     $todir=$main::TMPDIR;
     $l=$l." to $outlist"
   }   
   &main::mkLog($l);


# But... can I write to this directory ?!?  
       if (not -w $todir)  {
	 my $msg="Access dened to dir $todir !";
	 &main::mkLog($msg,"E",$step);
	 #$nerr++;
	 return 1;
       }
        

#////////////////////////////////////
# Check which steps should/may be done   
#////////////////////////////////////

   if ($main::DEBUG) {print "#DEBUG >>> Checking steps to be performed\n"}


# This variable we store the steps to be done (see sub getFlag for the definition)
   my $flag_step='';       

# Discard images by filter ?
   my $CH_FILTER=$main::params->{select}{xfilter}{Value};
    $CH_FILTER='no' if $main::params->{select}{filter}{Value} eq '';

   my $USEMASK = $main::params->{mask}{use}{Value} ;
   if ($USEMASK eq 'yes') {
    my $list= $main::params->{mask}{input}{Value};
    my $l="Images will be masked with the list: $list";      
    &main::mkLog($l); 
     if (! -e $list) {
          $USEMASK='no';   
	  my $msg="The list $list does not exist";
	  &main::mkLog($msg,"W",'Mask');
     }
   }
   
   my $DOOVERSCAN=$main::params->{overscan}{apply}{Value};
   if ($DOOVERSCAN eq 'yes' ) {
     #$nsteps++;
     $l="Overscan subtraction with: ";
     if ( $main::params->{overscan}{method}{Value} eq 'fit') {
       $l=$l." poly. of order $main::params->{overscan}{order}{Value}";
       $l=$l." ($main::params->{overscan}{mode}{Value})";
     }
     elsif ( $main::params->{overscan}{method}{Value} eq 'block') {
       $l=$l." block average, size:  $main::params->{overscan}{order}{Value}";
     } 
     &main::mkLog($l);
     $flag_step=$flag_step.getFlag('overscan');
   }
   
   my $DOTRIM =  $main::params->{trim}{apply}{Value};
   if ($DOTRIM eq 'yes' ) {
     $l="Trimming";
     #if ($nsteps==0) {$nsteps++};
     &main::mkLog($l);
     $flag_step=$flag_step.getFlag('trim');	
   }

   my $DOBIAS = $main::params->{bias}{apply}{Value};
   if ($step eq 'bias') { 
     $DOBIAS='no';
     $CH_FILTER='no';
   }
   if ($DOBIAS eq 'yes' ) {
     my $list= $main::params->{bias}{output}{Value};
     my $l="BIAS subtraction with list: $list";      
     &main::mkLog($l); 
     $flag_step=CheckStep($flag_step, 'bias', $list, \$DOBIAS);    
   }
   
   my $DODARK = $main::params->{dark}{apply}{Value};
   $DODARK='no' if ($step eq 'bias' || $step eq 'dark');
   $CH_FILTER='no'  if ($step eq 'dark');  
   if ($DODARK eq 'yes' ) {
     my $list= $main::params->{dark}{output}{Value};
     $l="DARK subtraction with list: $list";      
     &main::mkLog($l); 
     $flag_step=CheckStep($flag_step, 'dark', $list, \$DODARK);   
   }
   
   my $DODFLAT = $main::params->{dflat}{apply}{Value};
   $DODFLAT='no' if ($step eq 'bias' || $step eq 'dark' || $step eq 'dflat');
   
   if ($DODFLAT eq 'yes') {
     my $list= $main::params->{dflat}{output}{Value};
     my $l="DOME FLAT division with list: $list";      
     &main::mkLog($l); 
     $flag_step=CheckStep($flag_step, 'dflat', $list, \$DODFLAT);   
    }
   
   
   my $DOTFLAT = $main::params->{tflat}{apply}{Value};
   $DOTFLAT='no' if ($step eq 'bias' || $step eq 'dark' ||$step eq 'dflat' || 
              $step eq 'tflat');
   if ($DOTFLAT eq 'yes')  {
     my $list= $main::params->{tflat}{output}{Value};
     $l="TWILIGHT FLAT division with list: $list";      
     &main::mkLog($l); 
     $flag_step=CheckStep($flag_step, 'tflat', $list, \$DOTFLAT);   
     }
   
   my $f_bydate='no';   
   my $DOFRINGE = $main::params->{fringe}{apply}{Value};
   #$DOFRINGE='no' if ($step ne 'science');
   $DOFRINGE='no' if ($step eq 'bias' || $step eq 'dark' || $step eq 'dflat' || $step eq 'tflat' || $step eq 'fringe');
   if ($DOFRINGE eq 'yes' ) {
     my $list= $main::params->{fringe}{output}{Value};
     $l="FRINGE Scaling and subtraction with list: $list";      
     &main::mkLog($l); 
     if ($main::params->{fringe}{group}{Value} eq 'bydate') {
       $f_bydate='yes';
      $l="   Fringe frames are grouped by date";
       &main::mkLog($l);  
     }
     $flag_step=CheckStep($flag_step, 'fringe', $list, \$DOFRINGE);   
   }


   my $s_bydate="no";
   my $DOSFLAT = $main::params->{sflat}{apply}{Value};
   #$DOSFLAT='no' if ($step eq 'bias' || $step eq 'dark' || $step =~ /flat/);
    $DOSFLAT='no' if ($step ne 'science');
   if ($DOSFLAT eq 'yes' ) {
     my $list= $main::params->{sflat}{output}{Value};
     my $l="SUPER FLAT division with list: $list";      
     &main::mkLog($l);  
     if ($main::params->{sflat}{group}{Value} eq 'bydate') {
       $s_bydate='yes';
       $l="   Sflat frames are grouped by date";
       &main::mkLog($l);  
     } 
     $flag_step=CheckStep($flag_step, 'sflat', $list, \$DOSFLAT);    
   }
   

    my $DOWEIGHT='no';
    if ($step eq 'science') {
      $DOWEIGHT = $main::params->{science}{weight}{Value}; 
   if ($DOWEIGHT eq 'yes' ) {
     my $list= $main::params->{science}{wout}{Value};
     if ($main::redmode ne 'listonly') {
          system ("rm $list")  if (-e $list);
     }
     $l="Weight maps will be computed (a list $list will be produced)";      
     &main::mkLog($l);   
      }
    }

#=============== Selection rules..............
         
    if ($CH_FILTER ne 'no') {
      my $l="Images with \'FILTER NOT $CH_FILTER $main::params->{select}{filter}{Value}\' will be ";   
      if ($main::TFILTER==1) {$l=$l.' -skipped-'}
      else {$l=$l.' -reported-'} ;  
      &main::mkLog($l); 
    }
   
   
   if ($main::redmode eq "listonly") {return 0;}
   
   my $nerr=0;
   
   
   
   #////////////////////////////////////////////////
   # Make histogram of image (not very efficient)
   # A ps file will be created containing in each page the histogram of counts after each step,
   # one page for image; if specified, the fit of the overscan is also included
   # -----------------------------------------------
   my ($mmax, $mmin);
   my $nbins=500;
   my $objpsfile;
   
   if ($main::params->{science}{plot}{Value} eq 'yes' || $main::params->{overscan}{plot}{Value} eq 'yes') {
     my $nplots=0;
     #if ($main::params->{science}{plot}{Value} eq 'yes') {$nplots=$nsteps}
     #if ($main::params->{overscan}{plot}{Value} eq 'yes') {$nplots++}
     #$objpsfile="prered.ps";
     #dev "$objpsfile/PS", 2,  int($nplots/2+0.5)+1;
     #pgbegin(0,"$objpsfile/PS",2, int($nplots/2+0.5)+1);
     #pgsch(2.0);
   }
   #////////////////////////////////////////////////
   
   if (-e $outlist) {system ("rm $outlist")} #remove the output list, if it exists already
   
   
   if ($main::DEBUG) {print "#DEBUG >>> Preliminary steps\n"}
   
   if ($main::DEBUG) {print "#DEBUG >>> INSTRUMENT: $instrument\n"}
   my $numccd=$main::params->{NUMCCD};
   
   #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
   # PRELIMINARY STEPS (SHOULD BE MADE ONLY ONCE FOR ALL IMAGES
   #---------------------------------------------------------
   
   #these values are used only if the flat frames are not already normalized
   # ('NORM' keyword not found in the header)
   my $avg_dflat=1;
   my $avg_tflat=1;
   my $avg_sflat=1;


   #If masks are available, then trim and convert to bad pixels=1, good pixels=0 

#     my $mask=null;
#       my @mdims;		#the mask in 1D
#       my $errmask=0;

     my $gmask="";
       
       if  ($USEMASK eq 'yes') {
	 #Currently the mask file DOES NOT contain the chip number.........
	 #  $nmask=Match($gmask,$iccd);
	 # so...... just take the ith file in the list........
	 my $imask=$main::params->{mask}{input}{Value};
	 $imask =~ tr/@//d;
         
	 $gmask=$main::params->{mask}{output}{Value};
	 $gmask =~ tr/@//d;

         system ("rm $gmask") if (-e $gmask);

	 open (MDFILE, ">$gmask");

	 open (MFILE, $imask);

	 my $iccd=$main::params->{MINCCD};

print "Applying MASK\n";
       
         while (<MFILE>) {
                   
	     my $nmask=$_;
	     chomp($nmask);

	     my %hmsk;
	     tie ( %hmsk, 'Tie::IxHash');
	     %hmsk=&fitsio::rhead($nmask);

	     # Build the output filename and insert CCD number into the header
	     my $HEAD=Header->new();
	     $HEAD->{IMAGEID} = $iccd;
	     
	     my $fmask=$HEAD->buildName(\%hmsk, {STEP=>'mask', OUTDIR=>$main::STOREDIR});
	     
	     print MDFILE "$fmask\n";
           
          if ( $main::redmode eq "mklists") {$iccd++; next}

	 if (-e $fmask) {
	   if ($main::REDO eq 'yes' && $main::redmode eq 'RUN') {
	     system ("rm $fmask") 
	   }
	   else {$iccd++; next} 
	 }
  	 
	     $HEAD->Update(\%hmsk);

	 my ($mask, $ret)=&fitsio::rfits($nmask); #read the mask 
	 if ($main::DEBUG) {print "#DEBUG - Mask $nmask read $ret\n"};
	 	 
	 if ($DOTRIM eq 'yes') {
	   my @mdims=$mask->dims;


            my  @trim=$main::params->Section(step=>'trim', ccd=>$iccd);
  
            foreach my $i (0..3) {
               if ($trim[$i] < 0) {
                    $trim[$i] = $mdims[$i]-1 if ($i==1 || $i==3);
                    $trim[$i] = 0            if ($i==0 || $i==2);       
                } else {
                   $trim[$i]--;		#for images the origin is (0,0), not (1,1)
                }
              }  

	   $mask=$mask->slice("$trim[0]:$trim[1],$trim[2]:$trim[3]");
	 }
my @mdims=$mask->dims;

      # It should be: bad pixels = 1. good pixels = 0
	my $iavg=int(avg($mask)+0.5);
	 if ($iavg==1) {
	     $mask*=-1;
	     $mask+=1;
	     }
       
#write the mask as a byte
        my $wret=&fitsio::wfits(byte($mask), $fmask, \%hmsk);
	$iccd++;    
	   } #end of while


	 close (MFILE);
	 close (MDFILE);

       }
       



   if ($main::DEBUG) {print "#DEBUG Starting loop\n"}
   
   #////////////////////////////////////////////////
   #OK, HERE THE LOOP ON THE IMAGES STARTS !!!!!
   # Group all the files according to the same root
   #////////////////////////////////////////////////
   
   my $ibar=0;			#the counter for the progress bar
   my $procfiles=0;             #how many files were written to disk
   my $dfiles=0;                #how many files are already on disk   

   #for my $i (0..$#ifiles) {
   
   my @id=$main::params->CCDid();
   #my @imext=$main::params->CCDext();
   
    if ($ins eq 'esowfi' && $fits_type eq 'SIMPLE') {
            @id=(0..$#id);
     }

   if ($main::DEBUG) {my $nfiles=$#roots+1; print "#DEBUG will process $nfiles files ($numccd CCDs)\n"}
   
 ROOT: foreach my $root (@roots) {
     
     if ($fits_type eq 'MEF') {
       $ICLASS->Translate($root); #this will automatically read and "translate" the MEF fits header
     }
     
   LCCD:   for (my $pccd=0;$pccd<$numccd;$pccd++) { #loop on each CCD in sequence
       my $iccd=$id[$pccd];	# and get the CCD's ID
       
       if ($main::DEBUG) {print "#DEBUG ROOT: $root, ccd nr: $pccd/$numccd => ccd's ID: $iccd\n"}
       
             
       # Get the name of the INPUT file                      --> $pfile
       #   and (later on) build the name of the OUTPUT file  --> $newfile
       
       my $chip=sprintf("%02d", $iccd);
       
       my $pfile=$root;
       
       my $imext='';
       
       my @s=split(/\./,$pfile); #find extension
       my $ext=".$s[$#s]";
       my $newfile=$pfile;
       # we need to know the path of the file ($base)
       my ($base, $p, $s) = fileparse($newfile, $ext); 
       $base=~tr/\.//d;      #It must not contain dots...

       
       if ($fits_type eq 'SIMPLE') {
	 
	 #Add the CCD's ID to the root
	 my $s=$root.$chip;
	 # Take off the path
	 my @xs=split(/\//, $s);
	 my $xs=$xs[$#xs];
	 my $found=0;
	 
	 # Loop on the file list and extract the correct file
         #   when found, that file is removed from the array
	 my $if=-1;
	 foreach my $pf (@ifiles) {

	   $if++;
           chomp ($pf);

	   my $i=rindex($pf, $xs);



	   if ($i>=0) {$pfile=$pf; $found=1; splice (@ifiles, $if, 1); last}
	 }     
 
	 if ($found ==0) {
	   my $msg="File $xs not found in list ";
	   &main::mkLog($msg,"W",$step); 
	   next LCCD; 
	 }
       }			#end if FITS=SIMPLE
       elsif ($fits_type eq 'MEF') {
	 #the filename is xxxxx.fits[#ext], where #ext=1(=HDU) + running CCD  
	 $imext='['.($pccd+1).']';  
	 $pfile=$root.'.fits';
       }

      # --- check if the input file exists
       if (!-e $pfile && !-l $pfile)  {
	 my $msg="could not open input image $pfile ";
	 &main::mkLog($msg,"W",$step);
	 next LCCD;
       }
       
       
       # ======= HEADER initialization  
       
       my %h;
       tie (%h, 'Tie::IxHash');	#the header of the output image, order must be preserved
       %h=&fitsio::rhead($pfile.$imext);
       
       if ($fits_type eq 'SIMPLE') {
	 $ICLASS->Translate(\%h); #this will automatically read and "translate" the SIMPLE fits header
         $ICLASS->setCCD(\%h)   if ($ins eq 'esowfi');
 
      }                     
       elsif ($fits_type eq 'MEF') {$ICLASS->setCCD(\%h)}
      

       # -------- update the header with changes in keywords
       $ICLASS->Update(\%h); 

       
       # ======= SELECTION OF FILES
       #  === Currently implemented: FILTER
       
       my $ifilter=$h{FILTER};
       
       if ($CH_FILTER ne 'no') {
	 my $wfilt=CheckFilter($CH_FILTER, $ifilter, $main::params->{select}{filter}{Value} );
	   
	   if ($wfilt==1) {
	     my $msg="Image $pfile: mismatch in -FILTER => $ifilter-  NOT \`$CH_FILTER $main::params->{select}{filter}{Value} \'";	
	     if ($main::TFILTER==1) {
	       &main::mkLog($msg,"E",$step);
	       next ROOT;
	     } else {
	       &main::mkLog($msg,"W",$step);
	     }
	   }
       }


      # ---- now build the output filename
       $newfile=$ICLASS->buildName({NAME=>$base, FLAGS=>$flag_step, OUTDIR=>$todir});
              
       if ($main::DEBUG) {print "#DEBUG >>> $step Will process file: $pfile -> $newfile\n"}      
       if ($newfile !~ /\w+/) {last}


#append to the list the output name

           open (OFILE, ">>$outlist");
             print OFILE "$newfile \t $h{LOCDATE}\n";
           close (OFILE);

# - Update the counter and print the name of the image currently processed
       $ibar++;
       my $pc=int(100*$ibar/($nccd*($#roots+1))+0.5);
       my @fs=split('/', $pfile);
       if ($main::interface eq 'gui') {
	 print "counter $pc $fs[-1]$imext\n"; 
       }  else{ printf "Processed files: %.0f%s (%s%s)\n",$pc,'%',$fs[-1],$imext}


my $done=0;    #is the processing already done ??

#print "Applying wait\n";

#stop if we only want to build the lists of files 

       if ($main::redmode eq "mklists") {
	   $done=1
	}
              
       # --- check if output file already exists
       # If so, then delete it:
       #   - if redo everything OR
       #   - the creation date of one of the previous step is greater
       if (-e $newfile)  {
	 if ($main::REDO eq 'yes') {
	   my $msg="output file $newfile exists, deleting";
	   &main::mkLog($msg,"W",$step);
	   system ("rm $newfile");
	 } 
      	 else {		
	   my $msg="output file $newfile exists, skipping";
	   &main::mkLog($msg,"W",$step);  
	   $dfiles++;
	   $done=1;
	 }
       }		

       if (!$done) {

#Here it is a good place to check if previous steps exist, if $step is sflat or science
 my $pr_file=0;
     if ($step eq 'science' || $step eq 'sflat') { 
 my $pr_step=substr($flag_step,0,-1);
    $pr_file=$ICLASS->buildName({NAME=>$base, FLAGS=>$pr_step, OUTDIR=>$main::TMPDIR});
 if (-e $pr_file) {
    $pfile=$pr_file;  $imext=""; %h=&fitsio::rhead($pr_file);
    delete $h{STAT}; delete $h{OUTNAME}; $pr_file=1;
 } else {$pr_file=0}
} 

    
       # ------- STORE SOME ADDITIONAL INFOs...
       $h{$main::SOFT->{NAME}[0]} = "$main::SOFT->{NAME}[1]";
       $h{COMMENTS}{$main::SOFT->{NAME}[0]} = $main::SOFT->{NAME}[2];
       
       $h{$main::SOFT->{VERSION}[0]} = "'$main::SOFT->{VERSION}[1]'";
       $h{COMMENTS}{$main::SOFT->{VERSION}[0]} = $main::SOFT->{VERSION}[2];
       
       $h{$main::SOFT->{DATE}[0]} = $main::SOFT->{DATE}[1];
       $h{COMMENTS}{$main::SOFT->{DATE}[0]} = $main::SOFT->{DATE}[2];
       
       $h{$main::SOFT->{AUTHOR}[0]} = $main::SOFT->{AUTHOR}[1];
       $h{COMMENTS}{$main::SOFT->{AUTHOR}[0]} = $main::SOFT->{AUTHOR}[2];
       
       $h{$main::SOFT->{INST}[0]} =$main::SOFT->{INST}[1] ;
       $h{COMMENTS}{$main::SOFT->{INST}[0]} =$main::SOFT->{INST}[2] ;

   
              
# ================ OK, here the whole image is read ! 
       my ($image, $ret)=&fitsio::rfits($pfile.$imext);
       if ($main::DEBUG) {print "#DEBUG IMAGE $pfile$imext READ $ret\n"}       
       
       if ($image->isempty) {my $msg="$pfile - $ret";
			     &main::mkLog($msg,"E",$step);
			     $nerr++;
			     next LCCD;
			   }
       
       my @idims=$image->dims;

       
       #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
       # Diagnostic plot (to be changed)
       #my $objpsfile=$pfile;
       
       if ($main::params->{science}{plot}{Value} =~ /y/i) {
	 my $im1d=$image->clump(2);
	 my ($mean,$rms,$median,$min,$max) = stats($im1d);
	 my ($n, $p, $s) = fileparse($pfile, '.fits');
	 #$objpsfile=$p.$n.'H.ps';
	 $mmax=$max;
	 $mmin=$min;
	 (my $xvals, my $hist)=mkhist($im1d, $min, $max, $nbins);
	 my $xh=which($hist>1);
	 bin $xvals->index($xh), log10($hist->index($xh));
	 pglabel("Counts","Log N","$n: RAW") ;
	 $s=sprintf "median: %.2f, rms: %.2f",$median,$rms;	
	 pgsch(1.5);
	 pgmtxt('T',.5, .1, 0.0, $s);
	 pgsch(2.0);
	 pgnext();
	 
       }
       #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
       
       # =========================================================
       # =========== OK, START PROCESSING OF THE FRAME ===========
       # =========================================================
       
       my @trim=();
       my @ovscan=();
       
       if ($DOTRIM eq 'yes' || $DOOVERSCAN eq 'yes' )  {
	 #WARNING: The sections are stored sequentially, so we have to use >> $pccd <<
	 
 if ($DOTRIM eq 'yes') {
   @trim=$main::params->Section(step=>'trim', ccd=>$pccd);
    if ($main::DEBUG) {
       print "#DEBUG Trim section for CCD $iccd (ID): [$trim[0]:$trim[1],$trim[2]:$trim[3]]\n"
     }
   foreach my $i (0..3) {
      if ($trim[$i] < 0) {
        $trim[$i] = $idims[$i]-1 if ($i==1 || $i==3);
        $trim[$i] = 0            if ($i==0 || $i==2);       
      } else {
        $trim[$i]--;		#for images the origin is (0,0), not (1,1)
     }
    }  
 }

 if ($DOOVERSCAN eq 'yes') {
   @ovscan=$main::params->Section(step=>'overscan', ccd=>$pccd);
   if ($main::DEBUG) {
     print "#DEBUG Overscan section for CCD $iccd (ID): [$ovscan[0]:$ovscan[1],$ovscan[2]:$ovscan[3]]\n";
   }
   foreach my $i (0..3) {
     if ($ovscan[$i] < 0) {
       $ovscan[$i] = $idims[$i]-1 if ($i==1 || $i==3);
       $ovscan[$i] = 0            if ($i==0 || $i==2); 
     } else {
     $ovscan[$i]--;
   }
   }
 }
 
}
       
       my $mask=null;
       my $nmask="";
#       my @mdims;		#the mask in 1D
#       my $errmask=0;
       
       if  ($USEMASK eq 'yes') {
	    $nmask=Match($gmask,$chip);
	    if (! -e $nmask) {my $msg="MASK - $nmask does not exists";
			      &main::mkLog($msg,"E",$step);					                          $nmask="";
				       } else {
	        my $ret;
                ($mask, $ret)=&fitsio::rfits($nmask,-1,-1,'byte');

		my @nmask=split('/', $nmask);
                $h{BPM} = $nmask[-1];
	      }
       }
       
       
       my $PROCSTATS=0;		#set this flag to store the statistics
       
        my ($mean, $rms,$skew, $kurt, $min, $max);
       if (! $pr_file) {
	 #add the statistical information: mean, rms, skewness, kurtosis
	     # UNLESS an already partially prereduced file is read
	 ($mean, $rms,$skew, $kurt, $min, $max)=moments($image->clump(-1));
	 my $comm=sprintf "Mean %.3f RMS %.2f Skew %.3f Kurt %.3f",$mean, $rms,$skew, $kurt;
	 
	 chomp ($comm);
	 my $stat='STAT_R';
	 # my $stat='STAT_'.getFlag('bias');
	 
	 $h{$stat}=$comm;
       }       
       
       if ($mean > $main::params->{select}{highrej}{Value} ) {
	 my $msg=sprintf("Image %s is saturated (avg >= %d)",$pfile,$main::params->{select}{highrej}{Value});
	 &main::mkLog($msg,"W",$step);
	 $h{SATUR} = 1;
       }
       
       
       
       if ($DOOVERSCAN eq 'yes' ) {
	 
	 if ($main::DEBUG) {print "#DEBUG - Overscan subtraction\n"}
	 
	 if ($h{OVERSCAN}) {my $msg="Overscan already subtracted from $pfile ";
			    &main::mkLog($msg,"W",$step);
			  } else {

print "Applying overscan\n";
			    
			    my ($os, $x, $fit);

			    if ($main::params->{overscan}{method}{Value} eq 'fit' ) {			      
			      #a. fit with a polynomial of the given order
			      if ($main::params->{overscan}{mode}{Value} eq 'median') {
				#median over the os section
				$os=medover($image->slice("$ovscan[0]:$ovscan[1],$ovscan[2]:$ovscan[3]"));
			      } else {
				#average the os section
				$os=average($image->slice("$ovscan[0]:$ovscan[1],$ovscan[2]:$ovscan[3]"));
			      }
	      
			      my $osdim=$os->nelem;
			      my $ord=$main::params->{overscan}{order}{Value};
			      my $w=ones (short, $osdim);
			      $x=$os->xvals;
			      my ($coef, $rms)=polyfit1d($x, $os, $w, $ord+1);
			      $fit=float(polyeval($x, $coef)); #convert to FLOAT, otherwise it will be DOUBLE !!
			      #extrapolate the vector
			      if ($ovscan[2] != 0 || $ovscan[3] != ($idims[1]-1) ) {
				my $ffit=zeroes(float, $idims[1]);
				my $tmp;
				($tmp = $ffit->slice("$ovscan[2]:$ovscan[3]")) .= $fit;
				($tmp = $ffit->slice("0:$ovscan[2]")) .= $fit->at(0);
				my $d1=$idims[1]-1;
				($tmp = $ffit->slice("$ovscan[3]:$d1")) .= $fit->at(($fit->nelem)-1);
				$fit=$ffit;
			      }
    
			    }	else {                              
                              #b. compute median in blocks
			       $fit=zeroes(float,$idims[1]);
			       	$os=$image->slice("$ovscan[0]:$ovscan[1],$ovscan[2]:$ovscan[3]")->copy;
			      
			      my $sblock=$main::params->{overscan}{order}{Value};
			      if ($sblock==1) {
			        my $tfit=$fit->slice("$ovscan[2]:$ovscan[3]");
				my $mos;
				if ($main::params->{overscan}{mode}{Value} eq 'median') {
				    $mos=medover($os);
				 } else {$mos=average($os)}

				$tfit.=$mos;				

			      } else {
				
			      my $osmax=$ovscan[3]-$ovscan[2];
			                                   
			      for (my $ib=0; $ib<$osmax; $ib+=$sblock) {
				  my $i1=$ib; my $i2=$i1+$sblock;
                                  if ($i2>=$osmax) {$i2=$osmax};
				  my $osbl=$os->slice(",$i1:$i2");
				  my $i1f=$i1+$ovscan[2]; my $i2f=$i2+$ovscan[2];
				  my $tfit=$fit->slice("$i1f:$i2f");
				  my $mos=median($osbl); $tfit .= $mos;
				  #$osbl .= median($osbl);

			      }
			     }
			     
			      #extrapolate the vector
			      if ($ovscan[2] != 0 || $ovscan[3] != ($idims[1]-1) ) {
			      my $tmp;
 
			      ($tmp = $fit->slice("0:$ovscan[2]")) .= $fit->at($ovscan[2]);
			      my $d1=$idims[1]-1;
				($tmp = $fit->slice("$ovscan[3]:$d1")) .= $fit->at($ovscan[3]);
			      }
				 } #end if blocks	
	    

			      my $avfit=average($fit);
			    #subtract the fit from the image using backward assignment on a trasposed child piddle
			    my $imt=$image->mv(0,1);
			    $imt -= $fit;
			    
			    #(This would have been the same, but more memory consuming AND slower):
			    #subtract the fit from the transpose of the image
			    #my $imt=$image->mv(0,1) - $fit;
			    #and then take the inverse transpose
			    #$image=$imt->mv(1,0);
			    
			    my $today=&main::mkdate();
			    my $comm=sprintf "%s with [%d:%d,%d:%d], mean %.4f", $today, $ovscan[0]+1,$ovscan[1]+1,$ovscan[2]+1,$ovscan[3]+1, $avfit;
			    chomp ($comm);
			    $h{OVERSCAN}=$comm;
			    
			    #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
			    # Diagnostic plot (to be changed)
			    
			    if ($main::params->{overscan}{plot}{Value} =~ /y/i) {

                             my $psfile=$ICLASS->buildName({NAME=>$base, FLAGS=>'O', EXT=>'ps', OUTDIR=>$main::PSDIR});

			     my $osplot = dev("$psfile/CPS",1, 2);
			      my $xf = $fit->xvals;
			     my @osdims=$os->dims;
			     if ($#osdims >0) {
			     	 $os=medover($os);
			       }

			      $x=$os->xvals if (ref($x) ne 'PDL');
			      $x += $ovscan[2];

                              my $title = sprintf "%s: Overscan, created on %s, method %s/%d", 
$pfile.$imext,  $today, $main::params->{overscan}{method}{Value},
$main::params->{overscan}{order}{Value}  ; 

			      env (min($xf),max($xf),min($os),max($os), {CharSize=>1.8});
			     label_axes('Row', 'Counts', $title, {Colour=>'BLACK', 
							   AxisColour=>'BLACK', CHARSIZE=>1.4});

                              points ($x, $os,  {SYMBOL => 'DOT', CHARSIZE=>1.1});
			      line ($xf, $fit, {COLOUR=> 'RED', LINESTYLE=>'dashed'}); 


# === now plot the residuals; show them in the whole y range, but compute limits
#  only from the range where the fit was made

			      my $ros= medover($image->slice("$ovscan[0]:$ovscan[1]"));
 			      my $ios=which($xf>$ovscan[2] & $xf<$ovscan[3]);     

			      release;
			      env (min($xf),max($xf),minmax($ros->index($ios)), {CharSize=>1.8});
                                label_axes('Row', 'Residual', '', {AxisColour=>'BLACK', 
							   Colour=>'BLACK', CHARSIZE=>1.4});  
			 
                               line ($xf, $ros, {COLOUR=> 'BLACK'}) ;
			      #$osplot->close;

			      system("gzip -f $psfile");

			    }	#end plot overscan vector


			    #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
			    
			  }	#end if already overscan
       }			#end if do overscan
 

       if ($DOTRIM eq 'yes' ) {
	 
	 if ($main::DEBUG) {print "#DEBUG - Trimming\n"}

print "Applying trim\n";

	 if ($h{TRIM} && ($h{TRIM} =~ /\d+/)) {my $msg="Trimming of $pfile was already done.";
			&main::mkLog($msg,"W",$step);
		      } else {
			
			if ($main::DEBUG) {print "#DEBUG trim section: $trim[0]:$trim[1],$trim[2]:$trim[3]\n"};
			$image=$image->slice("$trim[0]:$trim[1],$trim[2]:$trim[3]");;
			#print " done !\n";
			
			my $today=&main::mkdate();
			my $comm=sprintf "%s with [%d:%d,%d:%d]", $today, $trim[0]+1,$trim[1]+1,$trim[2]+1,$trim[3]+1;
			chomp ($comm);
			$h{TRIM}=$comm;
			
			@idims=$image->dims; #the dimension CHANGED !!!!
			
		      }
       }
       

my $errmask=0;
       
       if ($USEMASK eq 'yes') {
	 #check that the dimension of the mask match with that of the image
          my @mdims=$mask->dims;
	 if ($mdims[0] != $idims[0] || $mdims[1] != $idims[1]) {


                 my $msg="The dimensions of mask ($nmask) and image ($pfile) do not match ";
		 &main::mkLog($msg,"E",$step);	
		 last;
                 $errmask=1;
               }
       }
       
       
       #ok, now start with frame operations
       
       
       if ($DOBIAS eq 'yes') {
	 
	 if ($main::DEBUG) {print "#DEBUG - Bias subtraction\n"}

print "Applying BIAS\n";


	 my $mkit=1;

	 if ($h{BIASCOR}) {
#                          my $bkw='F_IN'.getFlag('bias');
#                           my $bname=$h{$bkw};
#	                   my %bh=&fitsio::rhead($bname);
#			   my $bdate=&main::mkloctime($bh{PROCDATE});			   
#			   my $idate=&main::mkloctime($bh{BIASCOR});				  

#			   if ($bdate<$idate) {$mkit=0}
#			 }

#			   if ($mkit==0) {
                           my $msg="Bias was already subtracted from $pfile.";
			   &main::mkLog($msg,"W",$step);
			 }
                         else {
			   my $gbias=$main::params->{bias}{output}{Value};
			   my $nbias=Match($gbias,$chip);
			   if (! -e $nbias) {
			     my $msg="BIAS - the file $nbias does not exist ";
			     &main::mkLog($msg,"E",$step);
			     next LCCD;
			   } else {
			     my ($bias, $ret)=&fitsio::rfits($nbias);
			     if ($main::DEBUG) {print "#DEBUG   File $nbias read $ret\n"};
			     my @bdims=$bias->dims;
			     
			     if ($bdims[0] != $idims[0] || $bdims[1] != $idims[1]) {
			       my $msg="The dimensions of bias ($nbias) and image ($pfile) do not match";
			       &main::mkLog($msg,"E",$step);
			       last;
			     }
			     
			     CCDsubtract($image, $bias, $mask, $main::params->{mask}{badvalue}{Value});
			     
			     # if  ($USEMASK eq 'yes' && $errmask == 0) {
			     #            my $i_im=index($image->clump(-1), $m_idx);
			     #            my $i_b=index($bias->clump(-1), $m_idx);
			     #             $i_im -= $i_b;	
			     #  }
			     #   else {$image -= $bias}
			     
			     
			     my $today=&main::mkdate();
			     my $comm=sprintf "%s",$today;
			     chomp ($comm);
			     $h{BIASCOR}=$comm;
			     
			     my $datain='F_IN'.getFlag('bias');
			     my ($n, $p) = fileparse($nbias, '');
			     $h{$datain}=$n;
			     
			     if ($step eq 'science' ) {
			       #add the statistical information: mean, rms, skewness, kurtosis
			       ($mean, $rms,$skew, $kurt, $min, $max)=moments($image->clump(-1));
			       my $comm=sprintf "Mean %.3f RMS %.3f Skew %.3f Kurt %.3f",$mean, $rms,$skew, $kurt;
			       chomp ($comm);
			       
			       my $stat='STAT_'.getFlag('bias');
			       
			       $h{$stat}=$comm;
			       
			       $PROCSTATS=$comm;
			     }
			     
			   }	#end if file not found
			 }
       }
       
       if ($DODARK eq 'yes') {
	 
	 if ($main::DEBUG) {print "#DEBUG - Dark subtraction\n"}

print "Applying dark\n";
	 
	 if ($h{DARKCOR}) {my $msg="Dark was already subtracted from $pfile.";
			   &main::mkLog($msg,"W",$step);
			 } else {
			   my $gdark=$main::params->{dark}{output}{Value};
			   my $ndark=Match($gdark,$chip);
			   
			   if (! -e $ndark) {my $msg="DARK - $ndark does not exists";
					     &main::mkLog($msg,"E",$step);
					     next LCCD;
					   }
			   else {
			     my ($dark, $ret)=&fitsio::rfits($ndark);
			     if ($main::DEBUG) {print "#DEBUG   File $ndark read $ret\n"};
			     
			     my @ddims=$dark->dims;
			     
			     if ($ddims[0] != $idims[0] || $ddims[1] != $idims[1]) {
			       
			       my $msg="The dimensions of dark ($ndark) and image ($pfile) do not match";
			       &main::mkLog($msg,"E",$step);
			       #last;
			     }
			     
			     my %hdark=&fitsio::rhead($ndark);
			     my $SCALE=float($h{EXPTIME}/$hdark{EXPTIME});
			     
			     $dark *= $SCALE;
			     			     
			     CCDsubtract($image, $dark, $mask,  $main::params->{mask}{badvalue}{Value});
			     			     
			     my $today=&main::mkdate();
			     my $comm=sprintf "%s, scale %.4f",$today, $SCALE;
			     chomp ($comm);
			     $h{DARKCOR}=$comm;
			     
			     my $datain='F_IN'.getFlag('dark');
			     my ($n, $p) = fileparse($ndark, '');
			     $h{$datain}=$n;
			     
			     if ($step eq 'science' ) {
			       #add the statistical information: mean, rms, skewness, kurtosis
			       ($mean, $rms,$skew, $kurt, $min, $max)=moments($image->clump(-1));
			       
			       my $comm=sprintf "Mean %.3f RMS %.3f Skew %.3f Kurt %.3f",$mean, $rms,$skew, $kurt;
			       chomp ($comm);
			       
			       my $stat='STAT_'.getFlag('dark');
			       
			       $h{$stat}=$comm;
			       
			       $PROCSTATS=$comm;
			     }
			     
			   }	#end of check if file exists
			   
			 }	#end of check if darks was already subtracted
       }
       
       #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
       # Diagnostic plot (to be changed)
       
       if ($main::params->{science}{plot}{Value} =~ /y/i) {
	 # make the histogram after overscan, bias and/or dark subtractions
	 pgnext();
	 my $im1d=$image->clump(2);
	 my ($mean,$rms,$median,$min,$max) = stats($im1d);
	 $mmin=$min;
	 if ($max > $mmax) {$max=$mmax}
	 (my $xvals, my $hist)=mkhist($im1d, $min, $max, $nbins);
	 my $xh=which($hist>1);
	 bin $xvals->index($xh), log10($hist->index($xh));
	 pglabel("Counts","Log N","AFTER OS/BIAS/DARK SUBTRACTION") ;
	 my $s=sprintf "median: %.2f, rms: %.2f",$median,$rms;
	 pgsch(1.5);
	 pgmtxt('T',.5, .1, 0.0, $s);
	 pgsch(2.0);
       }			#end plot histo
       
       #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
       #   FLAT division
       #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\


         if ($DODFLAT eq 'yes') {
	 if ($h{DFLATCOR}) {my $msg="$pfile was already divided for dome flat";
			    &main::mkLog($msg,"W",$step);
			  }
	 else {
	   if ($main::DEBUG) {print "#DEBUG - DOME Flat division\n"}

print "Applying DFLAT\n";

	   my $rflat=DoFlat('dflat', $image, $pfile, \%h, \$avg_dflat, $mask, $step, \$PROCSTATS, $s_bydate);

          last if ($rflat==1);
 	 }
       }
       
       if ($DOTFLAT eq 'yes') {
	 if ($h{TFLATCOR}) {my $msg="$pfile was already divided for twilight flat";
			    &main::mkLog($msg,"W",$step);
			  }
	 else {
	   if ($main::DEBUG) {print "#DEBUG - TWILIGHT Flat division\n"}
print "Applying TFLAT\n";

	   my $rflat=DoFlat('tflat',  $image, $pfile, \%h, \$avg_tflat, $mask, $step, \$PROCSTATS);
          last if ($rflat==1);
	 }
       }
       

       if ($DOFRINGE eq 'yes' ) {
	 
	 if ($main::DEBUG) {print "#DEBUG - Fringe subtraction\n"}
print "Applying FRINGE\n";
	 
	 if ($h{FRINGCOR}) {my $msg="Fringes were already subtracted from $pfile";
			    &main::mkLog($msg,"W",$step);
			  } else {
			    my $fringe;
			    my $gfringe=$main::params->{fringe}{output}{Value};
			    my $nfringe;
    #if required, take from the list only the fringe frames matching by date
			    if ($f_bydate eq 'yes') {
			       my $ldate=$h{LOCDATE};
			       $nfringe=Match($gfringe,$chip, {KEY=>'LOCDATE', VALUE=>$ldate});
			     } else {
			        $nfringe=Match($gfringe,$chip);
			      }
			    if (! -e $nfringe) {
			      my $msg="Fringe - $nfringe does not exist ";
			      &main::mkLog($msg,"E",$step);
			     next LCCD;
			    } else {
			      if ($main::DEBUG) {print "#DEBUG Scaling to $nfringe\n"}
			      
			      my %frh;
			      
			      tie (%frh, 'Tie::IxHash');
			      %frh=&fitsio::rhead($nfringe);
			      my $ret;
			      ($fringe, $ret)=&fitsio::rfits($nfringe);
			      if ($main::DEBUG) {print "#DEBUG   File $nfringe read $ret\n"};
			      
			      my @fdims=$fringe->dims;
			      
			      if ($fdims[0] != $idims[0] || $fdims[1] != $idims[1]) {
				my $msg="The dimensions of fringe ($nfringe) and image ($pfile) do not match";
				&main::mkLog($msg,"E",$step);
				last;
			      }
			     
			      my $ffilter=$frh{FILTER};
			      my $wfilt=CheckFilter('cmp', $ifilter, $ffilter);

			      if ($wfilt) {
				my $msg="Image $pfile, mismatch in FILTER - $nfringe = $ffilter, image = $ifilter";  
				if ($main::TFILTER==1) {
				  &main::mkLog($msg,"E",$step);				  
				  next LCCD;
				}
				else {&main::mkLog($msg,"W",$step)};
			      #last;
			    }
			      
			      #  Fringe procedure  (Radovich & Bertin 24 January 2001):
			      #  Assume that the (bkg subtracted image) and master fringe are:
			      #    I = a + b F
			      # then evaluate a and b using a robust linear fit
			      # May 9 2001: objects detected with SEx are excluded from the fit

			      my $wdw=$main::params->{fringe}{window}{Value};
			      my @ws=split('/', $newfile);
			      my $nimsub=$main::TMPDIR.$ICLASS->buildName({NAME=>$ws[-1], FLAGS=>'sfrg'});
			      my $nimobj=$main::TMPDIR.$ICLASS->buildName({NAME=>$ws[-1], FLAGS=>'objs'});

			      if (! -e $nimsub || ! -e $nimobj) {
				my $tfile=$main::TMPDIR.'tmp.fits';
				my $ret= &fitsio::wfits($image, $tfile, \%h);
				if ($ret ne '') {next LCCD}
			#	if (! -e $nimsub) {
				  $ret=&main::doSEx($tfile, 'fringe', \$nimsub, \$nimobj);
			#	}
			#	if (! -e $nimobj) {
			#	   $ret=&main::doSEx($tfile, 'detect', \$nimobj);
			#	 }
				system ("rm $tfile");
			      } 
			      my ($imsub, $sret)=&fitsio::rfits($nimsub);
			      my ($imobj, $sret)=&fitsio::rfits($nimobj);

#warn "fringe: fitting $nimsub vs. $nfringe";
			    
			      my $im1da=$imsub->clump(-1);
			      my $fr1da=$fringe->clump(-1);
			      my $obj1da=$imobj->clump(-1);
			   			      			      
			      my $nz=which($im1da !=0 & $im1da > $main::params->{mask}{badvalue}{Value} & $obj1da==0);

			      my ($im1d, $fr1d);

			      if ($nz->isempty) {
				 $im1d=$im1da;
				 $fr1d=$fr1da;
			      } else {
				$im1d=$im1da->index($nz);
				$fr1d=$fr1da->index($nz);
			      }

			      my @ss=(); my $sig=0; my @cc=();

			      if ($wdw>1) {
			         my $xs=&Mylib::Robust::MedBlock($fr1d, $wdw);
			         my $ys=&Mylib::Robust::MedBlock($im1d, $wdw);
				 @cc=&Mylib::Robust::Robust_LinFit($xs, $ys, \$sig, \@ss);
			       } else {@cc=&Mylib::Robust::Robust_LinFit($fr1d, $im1d, \$sig, \@ss);}

			      $fringe*=$cc[1];
			      
			      CCDsubtract ($image, $fringe, $mask, $main::params->{mask}{badvalue}{Value} );
			      
			      # End of fringe scaling
			      
			      my $today=&main::mkdate();
			      my $comm=sprintf "%s, bkg %.2f, scale %.3f, ",$today, $cc[0], $cc[1];
			      chomp ($comm);
			      $h{FRINGCOR}=$comm;
			      
			      my $datain='F_IN'.getFlag('fringe');
			      my ($n, $p) = fileparse($nfringe, '');
			      $h{$datain}=$n;
			      
			      
			      if ($step eq 'science' ) {
				#add the statistical information: mean, rms, skewness, kurtosis
				($mean, $rms,$skew, $kurt, $min, $max)=moments($image->clump(-1));
				
				my $comm=sprintf "Mean %.3f RMS %.3f Skew %.3f Kurt %.3f",$mean, $rms,$skew, $kurt;
				chomp ($comm);
				
				my $stat='STAT_'.getFlag('fringe');
				
				$h{$stat}=$comm;
				
				$PROCSTATS=$comm;
			      }
			      
			      
			      #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
			      # Diagnostic plot (to be changed)
			      
			      if ($main::params->{science}{plot}{Value} eq 'yes') {
				# make the histogram after fringe correction
				pgnext();
				my $im1d=$image->clump(2);
				my ($mean,$rms,$median,$min,$max) = stats($im1d);
				
				if ($min < $mmin) {$min=$mmin}	
				if ($max > $mmax) {$max=$mmax}
				(my $xvals, my $hist)=mkhist($im1d, $min, $max, $nbins);
				my $xh=which($hist>1);
				bin $xvals->index($xh), log10($hist->index($xh));
				pglabel("Counts","Log N","AFTER FRINGE SUBTRACTION") ;
				my $s=sprintf "median: %.2f, rms: %.2f",$median,$rms;
				pgsch(1.5);
				pgmtxt('T',.5, .1, 0.0, $s);
				pgsch(2.0);
			      }	#end plot histo
			      #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
			    }	#end if file not found
			  }	#end else
       }			#end fringe correction
       


       if ($DOSFLAT eq 'yes') {
	 if ($h{SFLATCOR}) {my $msg="$pfile was already divided for super flat";
			    &main::mkLog($msg,"W",$step);
			  }
	 else {
	   if ($main::DEBUG) {print "#DEBUG - SUPER Flat division\n"}
print "Applying SFLAT\n";

	   my $rflat=DoFlat('sflat',  $image, $pfile, \%h, ,\$avg_sflat, $mask, $step, \$PROCSTATS);
          last if ($rflat==1);
	 }
       }
                   
       # =====> Now compute the FINAL stats
       #add the statistical information: mean, rms, skewness, kurtosis
     my $comm;
       if ($PROCSTATS==0) {
	 ($mean, $rms,$skew, $kurt, $min, $max)=moments($image->clump(-1));
	 $comm=sprintf "Mean %.3f RMS %.3f Skew %.3f Kurt %.3f",$mean, $rms,$skew, $kurt;
	 chomp ($comm);
       }  else {$comm=$PROCSTATS}
 
   my $stat='STAT';
   $h{$stat}=$comm;

  if ($image->info("%T") =~ /double/i) {
	     my $msg="Output image is double, converted to float";
	     &main::mkLog($msg,"W",$step); $image=float($image)
	 }


  my $df=&checkdf($image->info("%M"), $newfile);
  if ($df ne "") {&main::mkLog($df,"E",$step);$nerr++;return $nerr;}

#OK, write the image !!!!!!!!!

   my $today=&main::mkdate();
   $h{PROCDATE} = sprintf "%s", $today; 
   $h{COMMENTS}{PROCDATE} = 'The date when the file was written';

   my @ws=split('/', $newfile);
   $h{OUTNAME}=$ws[-1];
   $h{COMMENTS}{OUTNAME}='The name of the file after preprocessing'; 
   $main::_wfits=$newfile;  
      $ret= &fitsio::wfits($image, $newfile, \%h);
    $main::_wfits=''; 
   $procfiles++;


    if ($main::DEBUG) {
        print "#DEBUG   File $newfile WRITTEN $ret\n";
      }

    if ($ret ne "") {$ret="CFITSIO gives $ret"; 
		     &main::mkLog($ret,"E",$step);
		     $nerr++;
		     return $nerr;
		   }

} #end of BIG (if ! $done)
else {$procfiles++}

#======= Last,  produce the weight map, if requested
if ($step eq 'science' && $DOWEIGHT eq 'yes') {

  print "Applying wmaps\n";

  if ($done) {%h=&fitsio::rhead($newfile)}; # == the file did not went through proc.

       my $hflat="";
       my $hmask="";

# ==== weight maps are built using either dome or twilight flats
# ====     and BPM files

       my $kflat='';
       if (exists $h{DFLATCOR}) {$kflat='F_IN'.getFlag('dflat')}
       elsif (exists $h{TFLATCOR}) {$kflat='F_IN'.getFlag('tflat')}
       
    if ($kflat) {$hflat=$main::STOREDIR.$h{$kflat}; $hflat=~tr/'//d;}

    if (exists $h{BPM}) {$hmask= $main::STOREDIR.$h{BPM}; $hmask=~tr/'//d;}
   my $oweight=$main::STOREDIR.$ICLASS->buildName({NAME=>$base,FLAGS=>'WW'}); 
   # $h{WMAP}=$oweight;

my $wwret=-1;


  if ($main::redmode eq 'RUN') { 
    if (-e $oweight && $main::REDO eq 'yes') {system ("rm $oweight")};
    if (!-e $oweight) {
          $wwret=doWW({FLAT=> $hflat , BPM=> $hmask, SCIENCE=> $newfile, MKWEIGHT=>$oweight });
	}
 }

  if ($wwret==1 || $main::redmode eq 'mklists') {
      open (WWFILE, ">>$main::params->{science}{wout}{Value}");
      print WWFILE  "$oweight\n";
      close (WWFILE);
   }

}

 } #end loop on CCDs
} #end loop on roots


# ======  Keep track of the list 
  &main::SumList($outlist, 'processing', $step, $procfiles, $dfiles);



#\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
# Diagnostic plot (to be changed)

 if ($main::params->{science}{plot}{Value} eq 'yes' ) {
       pgend;
      my $gzfile=$objpsfile.".gz";
      if (-e $gzfile) {system("rm $gzfile")};
      system("gzip $objpsfile");
      }

#\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\


if ($main::DEBUG) {print "#DEBUG == CCDPROC END ==\n\n"}

return ($nerr, $procfiles);

} #end ccdproc


  sub DoFlat
    {
# divide the image by the given flat
# (piddles are passed by reference !)      
      my ($wflat, $image, $imname, $h, $avg_wflat, $mask, $step, $PROCSTATS, $bydate)=@_;

      my $chip=$$h{IMAGEID};

      my $ftype="";
      if ($wflat eq 'dflat') {$ftype = 'Dome'}
      elsif ($wflat eq 'tflat') {$ftype = 'Twilight'}
      elsif ($wflat eq 'sflat') {$ftype = 'Super'}
      
      my $gflat=$main::params->{$wflat}{output}{Value};    

#if required, take from the list only the flat frames matching by date
    my $nflat;

       if ($bydate eq 'yes') {
	      my $ldate=$$h{LOCDATE};
	      $nflat=Match($gflat,$chip, {KEY=>'LOCDATE', VALUE=>$ldate});
	  } else {
	      $nflat=Match($gflat,$chip);
	 }

      if (! -e $nflat) {
	my $msg="$ftype flat - $nflat does not exist ";
	&main::mkLog($msg,"E",$step);
	next LCCD;
      }
      
      my ($flat, $ret)=&fitsio::rfits($nflat);
      
      if ($main::DEBUG) {print "#DEBUG   File $nflat read $ret\n"};
      
      my @fdims=$flat->dims;
      my @idims=$image->dims;
     

      if ($fdims[0] != $idims[0] || 
                       $fdims[1] != $idims[1]) {
	my $msg="Image $imname: the dimensions of flat ($nflat) and image do not match";
	&main::mkLog($msg,"E",$step);
	return 1;
      }
      
      my %fh=&fitsio::rhead($nflat);

      # be sure that we are using the same chips and the same filters.....
      my $fchip=$fh{IMAGEID};
  
      if ($fchip != $chip) {
         my $msg="Image $imname: mismatch in IMAGEID - $nflat = $fchip, image = $chip";  
         &main::mkLog($msg,"E",$step);
	 return 1;
      }
      my $ffilter=$fh{FILTER};
      my $ifilter=$$h{FILTER};
  
     my $wfilt=CheckFilter('cmp', $ifilter, $ffilter);
     if ($wfilt) {
         my $msg="Image $imname: mismatch in FILTER - $nflat = $ffilter, image = $ifilter";  
	 if ($main::TFILTER==1) {
	      &main::mkLog($msg,"E",$step);
	      return 1;
	   }
	 else {&main::mkLog($msg,"W",$step)};
      }
      
      #Now... is the flat already normalized ??
      my $nfac=1;
      
      if (! exists $fh{NORM} ) {
	my $fmode=$main::params->{$wflat}{nmode}{Value};
	my $ustat=0;
	
	if ($$avg_wflat!=1) {	#do this just once
	  ($$avg_wflat, $ustat)=NormFlats(LIST=>$gflat, FLAT=>$wflat);
	}
	
	if ($fmode eq 'single') {
	  if ($ustat) {
	    %fh=&fitsio::rhead($nflat); #reread the header with the STAT info
	  }
	  my $stat=$fh{STAT};
	  my @s=split(/ /, $stat);
	  $nfac=$s[1];
	}
	elsif ($fmode  eq 'global') {$nfac=$$avg_wflat}
	if ($main::DEBUG) {print "#DEBUG Normalization factor: $nfac\n"}
	$flat /= $nfac;
      } else {$nfac=$fh{NORM}}
      
      CCDdivide($image, $flat, $mask,  $main::params->{mask}{badvalue}{Value});
        	   
      my $today=&main::mkdate();
      my $comm=sprintf "%s, norm %.4f",$today, $nfac;
      chomp ($comm);
      my $kcomm=uc($wflat).'COR';
      $$h{$kcomm}=$comm;       
      $$h{COMMENTS}{$kcomm} = "$ftype flat correction";
      
      my $datain='F_IN'.getFlag($wflat);
      my ($n, $p) = fileparse($nflat, '');
      $$h{$datain}=$n;
      
      
      if ($step eq 'science' ) {
	#add the statistical information: mean, rms, skewness, kurtosis
	my ($mean, $rms,$skew, $kurt, $min, $max)=moments($image->clump(-1));
	
	my $comm=sprintf "Mean %.3f RMS %.3f Skew %.3f Kurt %.3f",$mean, $rms,$skew, $kurt;
	chomp ($comm);
	my $stat='STAT_'.getFlag($wflat);
	$$h{$stat}=$comm;
	
	$$PROCSTATS=$comm;
      }
      
#      #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
#      # Diagnostic plot (to be changed)
      
#      if ($main::params->{science}{plot}{Value} eq 'yes') {
#	# make the histogram after dome flat division
#	pgnext();
#	my $im1d=$image->clump(2);
#	my ($mean,$rms,$median,$min,$max) = stats($im1d);
	
#	if ($min < $mmin) {$min=$mmin}
#	if ($max > $mmax) {$max=$mmax}
#	(my $xvals, my $hist)=mkhist($im1d, $min, $max, $nbins);
#	my $xh=which($hist>1);
#	bin $xvals->index($xh), log10($hist->index($xh));
#	pglabel("Counts","Log N","AFTER FLAT FIELD DIVISION") ;
#	my $s=sprintf "median: %.2f, rms: %.2f",$median,$rms;
#	pgsch(1.5);
#	pgmtxt('T',.5, .1, 0.0, $s);
#	pgsch(2.0);
#      }				#end plot histo
#      #\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
      
  return 0;
    }				#end sub doFlat
  

  sub CheckFilter 
    {
      my ($mode, $filt1, $filt2)=@_;
      
      $filt1=lc($filt1);
      $filt2=lc($filt2);

      my $warn=0;
      if ($mode eq 'cmp' ) {
	#compare two keywords and return '1' if they are not the same       
	$filt1=~tr/\'| //d; 
	$filt2=~tr/\'| //d;
	$warn=1 if ($filt1 ne $filt2);
      }
      else {
	#the same criteria as in 'sub troper' in Tools::mklist.pl
      $warn=1;
	if ($mode eq 'is') {
	  $warn=0 if $filt1 eq $filt2;
	}
	elsif ($mode eq 'contains') {
	  $warn=0 if  $filt1 =~ $filt2;
	}
	elsif ($mode eq 'begins_with') {
	  $warn=0 if  $filt1 =~ /^$filt2/;
	}
#	elsif ($mode eq 'ends with') {
#	  $warn=0 if  $filt1 =~ /$filt2$/;
#	}

      }
      return $warn;
    }				#end sub CheckFilter
  
  sub CheckStep
  {
    my ($flag_step, $lab, $list, $dostep)=@_;
     my $nflag_step=$flag_step;
    if (! -e $list) {
      $$dostep='no';
      my $msg="$lab correction should be done but the list $list does not exist";
      &main::mkLog($msg,"W",'');
    }
    else {    
      #if ($$nsteps==0) {$$nsteps++};     
      $nflag_step=$flag_step.getFlag($lab);
     }
    return $nflag_step;
} #end sub Checkstep
   
  
  sub NormFlats
    #compute/apply a normalization to a list of flats, check for saturation
#returns a list of the average for each image

{

my %user=@_;

my ($flist,$wflat, $donorm);
if (exists $user{LIST}) {$flist= $user{LIST}} else {return}
if (exists $user{FLAT}) {$wflat= $user{FLAT}} else {return}
if (exists $user{DONORM}) {$donorm=1} else {$donorm=0}

my $mode= $main::params->{$wflat}{nmode}{Value};
my $satur=$main::params->{select}{highrej}{Value};

my @outlist=();
open (FILE, "<$flist");
while (<FILE>) {
  chomp ($_);
  push @outlist,$_;
}
close (FILE);

my @avgs=();
my $avg_flat=0;
my $ustat=0;

my $n=0;

my %fh;
tie (%fh, 'Tie::IxHash');  #it will be written, so we must tie it to preserve order

foreach my $nflat (@outlist) {
    my $avg;
    %fh=&fitsio::rhead($nflat);
    if (exists $fh{NORM}) {$avg = $fh{NORM}} else {

      if (exists ($fh{STAT})) {
         my $stat=$fh{STAT};	
	 my @s=split(/ /, $stat);
         $avg= $s[1];
       } else {                                 #Nope, we must compute the average
if ($main::DEBUG) {print "#DEBUG STAT not found, reading $nflat\n"};	
           $ustat=1;
	   my ($iflat, $ret)=&fitsio::rfits($nflat);
	   my ($rms,$skew, $kurt, $min, $max);
	   ($avg, $rms,$skew, $kurt, $min, $max)=moments($iflat->clump(-1));
           my $comm=sprintf "Mean %.3f RMS %.3f Skew %.3f Kurt %.3f",$avg, $rms,$skew, $kurt;
            chomp ($comm);
            $fh{'STAT'}=$comm;
	    $ret=&fitsio::whead($nflat, \%fh);
	 }
    }  # end of if NORM not found
	 push @avgs, $avg;
	 if ($avg > $satur) {next}
	 if ($mode eq 'global') {$avg_flat += $avg; $n++}	

 } #end loop on images

     if ($n >0 && $mode eq 'global') {
         $avg_flat/=$n;
         if ($main::DEBUG) {print "#DEBUG NormFlats: average of averages = $avg_flat ($n CCDs)\n";}
      }

   if ($donorm) {
    my $i=-1;
    my $nfac=1;
    FLAT: foreach my $name (@outlist) {
       $i++;
       %fh=&fitsio::rhead($name);
       next FLAT if (exists $fh{NORM});
       my ($image, $ret) = &fitsio::rfits($name);
       if ($avgs[$i]>$satur ) {
          $fh{SATUR}=1;
           my $msg=sprintf("uc($wflat) %s, average (%.0f) >  saturation limit (%d)",
                 $name, $avgs[$i]);
           &main::mkLog($msg,"W",'NormFlats')
        }
       if ($mode eq 'global') {$nfac=$avg_flat}
       elsif ($mode eq 'single') {$nfac=$avgs[$i]}
       else {$mode = 'none'}
        if ($nfac !=0 && $nfac !=1) {
	    $image/=$nfac;
            $fh{NORM}=sprintf("%.3f",$nfac);
            $fh{COMMENTS}{NORM}="The value used for the normalization";
          }
       $fh{NMODE}=$mode;
       $fh{COMMENTS}{NMODE}="The normalization mode [global/simple/none]";

       system ("rm $name");
       $main::_wfits=$name;  
          &fitsio::wfits ($image, $name, \%fh);
       $main::_wfits=''; 
     } #end of loop
   } #end of (if $donorm)

return ($avg_flat, $ustat);

} #end sub NormFlats



sub Match {

my $opt; $opt = pop @_ if ref($_[$#_]) eq 'HASH';

my ($root, $iccd)= @_;

my $found="";

my @files;

#if ($root =~ /\A@/) {        #a LIST of files is given
   my $mroot= $root;
   $mroot =~ tr/@//d;
   open (IFILE, "$mroot");
       while (<IFILE>) {chomp($_); push(@files,$_)}
   close (IFILE);
#   }
#else {my $mroot=$root.'*';@files=glob($mroot)}  #the ROOT is given (e.g. Bias*)

my $ncrit=1;

my ($key, $val);
if (exists $opt->{KEY} && exists $opt->{VALUE}) {$key=$opt->{KEY}; $val=$opt->{VALUE}; $ncrit++}


 foreach my $file (@files) {
   my $ok=0;
     my %hfr=&fitsio::rhead($file);
     my $ccd=$hfr{IMAGEID};    #the compulsory one....
     if ($ccd == $iccd) {$ok++}
     if ($ncrit >1) {
        my $ival=$hfr{$key};
	$ival=~tr/'//d;
        $val=~tr/'//d;
        $ok++ if ($val eq $ival);
     }

     if ($ok == $ncrit) {$found=$file; last}
 }

if ($found eq '') {my $msg="Nothing found in list $root !"; &main::mkLog($msg,"E",'match');}

return $found;

} #end sub Match


sub checkmem

{
my @top=top;
my @a=grep(/Memory/,@top);   #capture output from top
@a=split (/Real:/,$a[0],2);  #get the 'memory' line
@a=split (/M/,$a[1]);        #Take act & tot memory
my $mem=$a[1];               #we need the tot field
$mem =~ tr/\///d;


return $mem;
}




sub checkdf

{
#takes as argument the output of $pdl->info("%M"), which gives the memory in Mb taken by a pdl
# and the file to be created
my $info=shift;
my $to=shift;

# if we did not get an absolute path, then take the "local" disk
if (substr($to,0,1) ne '/') {$to=cwd}

my $msg="";

my @arr=split(/mb/i,$info,2);

my $d=$arr[0]*1024;

my @s=split(/\//, $to);
my $p='/'.$s[1];

#warn "$info -> $to -> $p";

#my ($n, $p)=fileparse($to);

my @df=df("-k",$p);

my $t=$df[1];   #get second line
$t =~ tr/ //s;   #get rid of repeated spaces
@arr=split(/ /,$t,6);   #there should be 6 entries, we want the 4th which gives the available space in kb

if ($#arr <3) {warn "CheckDF could not get the space available on $p !"}
else {
if ($d > $arr[3]) {
     $msg="Not enough space on >$p< to write the image !!!!!!";
     print "Image is $d kb, only $arr[3] kb are available !!\n";
     }
}
return $msg;

}


sub smFrame
{
#////////// SMOOTHING OF FRAMES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\
# $gframe => list of images
# $swdw => boxsize

my ($gframe, $swdw)=@_;
#my $swdw=$main::params->{frame}{smooth}{Value};

warn "#TEST smFrame:  $gframe";

   my $l="Smoothing frames from list $gframe ";
   if (! -e $gframe) { $l=$l."  (NOT FOUND !!!)"};
   $l=$l.", window size: $swdw pixels";

   &main::mkLog($l);

   return if ($main::redmode eq "listonly");

  if (! open (FILE, $gframe)) {return -1}

  my $tl="_tmp.lst";
  open (OFILE, ">$tl");

  while (<FILE>) {
     chomp ($_);
     my $nframe=$_;
warn "#TEST $nframe";
     my @s=split(/\./, $nframe);
  #Build the output name
     $s[0]=sprintf("%s_s%02d",$s[0],$swdw);     
     my $sfrm=join(".", @s);
#     foreach my $i (1..$#s) {$sfrm=$sfrm.'.'.$s[$i]}
warn "#TEST $sfrm";

     if (-e $sfrm) {
       if ($main::REDO eq 'yes') {next}
       else {system ("rm $sfrm")}
     }

if ($main::DEBUG) {print "#DEBUG Smooth frame $nframe to $sfrm \n"}
     my ($frame, $ret)=&fitsio::rfits($nframe);
     if ($main::DEBUG) {print "#DEBUG smoothing frame\n";}
       $frame= float(med2d $frame, ones($swdw,$swdw), {Boundary => 'Reflect'}) ;
   #now save it
      my %hsf;
      tie ( %hsf, 'Tie::IxHash');
      %hsf=&fitsio::rhead($nframe);
      $main::_wfits=$sfrm;  
         my $wret=&fitsio::wfits($frame, $sfrm, \%hsf);
      $main::_wfits=''; 
      print "#DEBUG $wret: $sfrm saved \n" if ($main::DEBUG) ;
      print OFILE "$sfrm\n";
 }
 close (FILE);
 close (OFILE);

  my $ns_fr=$gframe.'ns';
 system ("cp $gframe $ns_fr");
 system ("mv $tl $gframe");

} #end sub smFrame


sub mkhist {
#modified version of the PDL HIST to use number of bins instead of the width
    barf('Usage: ([$xvals],$hist) = myhist($data,[$min,$max,$bins])') if $#_<0;
    my($pdl,$min,$max,$bins)=@_;
    $min = $pdl->min() unless defined $min;
    $max = $pdl->max() unless defined $max;
    $bins=512 unless defined $bins;
    my $ntype = $pdl->get_datatype;

    my $step=($max-$min)/$bins;

    PDL::Primitive::histogram($pdl->clump(-1),(my $hist = null),
                              $step,$min,$bins);
    my $xvals = $min + $step/2 + sequence(PDL::Type->new($ntype),$bins)*
        PDL::convert($step,$ntype) if wantarray();
    return wantarray() ? ($xvals,$hist) : $hist;
}


sub getFlag
{
my $step=shift;

my $flag='';

if ($step eq 'overscan') {$flag='O'}
elsif ($step eq 'trim') {$flag='T'}
elsif ($step eq 'bias') {$flag='B'}
elsif ($step eq 'dark') {$flag='D'}
elsif ($step eq 'dflat') {$flag='Y'}
elsif ($step eq 'tflat') {$flag='W'}
elsif ($step eq 'sflat') {$flag='F'}
elsif ($step eq 'fringe') {$flag='G'}

return ($flag);
 
} #end sub getFlag



1;
