
use PDL;
use CFITSIO qw( :longnames :constants);
use PGPLOT;
use PDL::Fit::Polynomial;
use PDL::Image2D;
use  PDL::Graphics::PGPLOT;
use strict;

CFITSIO::PerlyUnpacking(0);

my @list=("/u2/mellier/Preredtest/CFH12K/CFH12K/99IIF156/B/1999-11-12/TWFLATB55s/55s/508591f/508591f00.fits");
#my @list=("t.fits");
#my @list=("/i2/radovich/data/pdl/Zerob00.fits");
#my @list=("/i2/radovich/data/may99/CFH12K/F31/BIAS/1999-05-08/0s/485886b/485886b00.fits");
#my @list=("del.fits");

         #Create a stack of images
	  my ($image, $ret) = rfits($list[0]);    #read the first image of the list
            if ($image->isempty) {print "$ret\n"}
	    else {print	"Tav: ", average($image->clump(2)),"\n"; }

	 #Now check how much memory it is taking
         # Each image of dimension ($nx,$ny) will be splitted in ($blocks) blocks of dimension ($deltax,$ny)

	  my $sp=$image->info("%M");
	  my @arr=split(/mb/i,$sp,2);
	  $sp=($#list+1)*$arr[0];   #total space taken in memory by all images
	  
	  my ($nx, $ny) = $image->dims;    #Image dimension



my $maxdim=200;
$maxdim /=4.;      
	  	
	     #$blocks=$sp/$maxdim; #e.g. 60 images of 32Mb => each may take 1Gb/60=16.7Mb, that is should be splitted in 1.92 parts
	     #$blocks=int($blocks+1);  #take the next integer 
my $blocks=10;
             my $ncim= zeroes($image->type,$nx,$ny); 

        my $deltax=int($nx/$blocks+0.5);  #This will be the width of each block

if ($blocks > 1) {print "WARNING: storing all images would take $sp Mb, maximum = $maxdim Mb: they will be splitted in $blocks blocks, nx = $deltax/$nx, ny=$ny\n";}

       my $cim; #this will be a huge array, make sure that it is deallocated	
       my $scalf=zeroes($image->type,$#list+1);   # This array will contain the scaling factors

      for my $nblock (1..$blocks) {	    
	   	 
	  my @xsec=( 1+($nblock-1)*$deltax, $nblock*$deltax);  # The x-section to be read
	  if ($xsec[1] > $nx) {$xsec[1] = $nx}       # the last block is usually smaller than the others
	  my $secnx=$xsec[1]-$xsec[0]+1;             # Actual size of each block in x
 	  my $sec="[".$xsec[0].":". $xsec[1].",1:$ny]";  # OK, this is the section to be read
         # my $sec="[$xsec[0]:$xsec[1],*]";
	#my $sec="";	 		
          my $stack= zeroes($image->type,$secnx,$ny,$#list+1);  #build a stack for all blocks          	
	
          # now fill $stack with the image blocks in the list	  	
          for my $k (0..$#list) {

	       my $filesec=$list[$k].$sec;    # Filename+section to be read
	

            #  if ($k > 0 || $blocks > 1) {
                   ($image, $ret)=rfits($filesec,$secnx,$ny); # Read the section, we must give also its size
            if ($image->isempty) {print "$ret\n"; die}
print "P av: ", average($image->clump(2)),"\n";
print "$sec $secnx $ny\n";
           #   }
	 }     

}

sub getheadtype
{
#get the right type for the values in the hader (called by wfits)
#the code was borrowed from the PDL IO library

my $val=shift;

my $typ;

 if ($$val =~ /^ *([+-]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))? *$/) { # Number?
        $typ=TFLOAT;
	$typ=TSHORT if ($$val !~ /\./) ;	#Is it really a float ?
  }  elsif ($$val eq 'F' or $val eq 'T') {          # Logical flags ?
      $typ=TLOGICAL;
      
} else {
         $$val=$1 if ($$val=~m/'(.*)'/); # Take off surrounding ''
          $typ=TSTRING;
  }
  
 return $typ;
}  #end sub getheadtype	  

sub rfits
{

my ($file, $ncols, $nrows)= @_;


my $ret="";

my $status = 0;
my $fptr = CFITSIO::open_file($file, READONLY, $status);

if ($status ne 0) {CFITSIO::fits_get_errstatus($status,$ret); return $ret}

my ($naxis1, $naxis2, $bitpix, $bscale, $bzero);

$fptr->read_key(TLONG,'BITPIX',$bitpix,undef,$status);
if ($status ne 0) {CFITSIO::fits_get_errstatus($status,$ret); $fptr->close_file($status);return $ret}

if (! $ncols) { 
   $fptr->read_key(TSTRING,'NAXIS1',$naxis1,undef,$status);
   if ($status ne 0) {CFITSIO::fits_get_errstatus($status,$ret); $fptr->close_file($status);return $ret}
} else {$naxis1=$ncols}

if (! $nrows) { 
   $fptr->read_key(TSTRING,'NAXIS2',$naxis2,undef,$status);
   if ($status ne 0) {CFITSIO::fits_get_errstatus($status,$ret); $fptr->close_file($status);return $ret}
} else {$naxis2=$nrows}

$fptr->read_key(TLONG,'BSCALE',$bscale,undef,$status);
$bscale="" if ($status ne 0);
$fptr->read_key(TLONG,'BZERO',$bzero,undef,$status);
$bzero="" if ($status ne 0);

$status=0;

my %funcs = (
        '8' => { 'pdl' => \&byte, 'cfitsio' => \&CFITSIO::fits_read_2d_byt },
        '16' => { 'pdl' => \&short , 'cfitsio' => \&CFITSIO::fits_read_2d_sht },
        '32' => { 'pdl' => \&long, 'cfitsio' => \&CFITSIO::fits_read_2d_lng },
        '-32' => { 'pdl' => \&float , 'cfitsio' => \&CFITSIO::fits_read_2d_flt },
        '64' => { 'pdl' => \&double , 'cfitsio' => \&CFITSIO::fits_read_2d_dbl },
);

#Handle unsigned short 
if ($bscale != "" && $bzero != "" && $bitpix==16) { 
  if (! $ncols && ! $nrows) {
    $funcs{$bitpix}{'pdl'}=\&ushort;
    $funcs{$bitpix}{'cfitsio'}=\&CFITSIO::fits_read_2d_usht;
    $bscale="";
    $bzero="";
  }  
}

my ($pdl,$anynul);
if (exists $funcs{$bitpix})
{
    $pdl = &{$funcs{$bitpix}{'pdl'}} (zeroes($naxis1,$naxis2));
    &{$funcs{$bitpix}{'cfitsio'}}($fptr,1,0,$naxis1,$naxis1,$naxis2,${$pdl->get_dataref},$anynul,$status);
    if ($status ne 0) {CFITSIO::fits_get_errstatus($status,$ret); $fptr->close_file($status);return (null,$ret)}
    $pdl->upd_data;
    $fptr->close_file($status);
    $pdl=float($pdl);  #convert piddle to float
    if ($bscale !="" && $bzero != "") { $pdl *= $bscale; $pdl += $bzero}
}
else
{
    $fptr->close_file($status);
    $ret= "$0: invalid BITPIX keyword (= $bitpix) in image '$file'";
}
print "rfits $bitpix $naxis1 $naxis2 : ", average($pdl->clump(2)),"\n";    
	
 return ($pdl, $ret);

} #end sub rfits

