use PDL;
use PDL::Math;

use Tk::BrowseEntry;

use vars qw/@header @frames $dbfile $odbfile @log @newtab/;
my $DEBUG=0;

sub checkdb {

my $tol;
($dbfile, $odbfile, $tol)=@_;



my $thresh= float [$tol/60./15., $tol/60.];   #convert from arcmin to hours/degrees
#High values => few groups 
#Low values =>  many groups

if (!open (DBFILE, $dbfile)) {	$main::top->Dialog(
            -title  => 'ERROR !!!',
            -text   => "Can't open file $dbfile",
            -bitmap => 'error',
        )->Show;
	return;
     };

@log=();
while (<DBFILE>) {
   chomp($_);
   push (@log, $_);
}
close (DBFILE);


#Note: it is more convenient to keep the entries of  @flog as strings, not as arrays,
# since so we can use grep to filter the array

#Remove from the array the header and the entries for dark, bias or flat

my $row;

$row=$log[0];
  @header=split(/$main::DBDEL/,$row);
my @flog=grep(!/$row|bias|dark|flat|dome|^\s$/i,@log);


#STEP 1: make subsample by names

my @onames=();
my @xflog=@flog;

my $nf=0;

# This loop will:
# read and split the first row of @xflog
# store the content 'object' field into the @onames array
# remove all rows in @xflog with the same name in 'object'
# stop when @xflog is empty
# result: the @onames will contain all names in the db.log as unique entries

while(@xflog)  {
 my @fields=split(/$main::DBDEL/,$xflog[0]);
   my $xname;
   my $c=0;
   foreach my $field (@fields) {
      chomp ($field);
      if ($header[$c] =~ "object" ) {$field=~ tr/ //d; $xname=$field;push(@onames,lc($xname))}
      $c++;
    }
   $nf=grep(/\b$xname\b/i,@xflog);
 if ($DEBUG){ print "-> #$xname# $nf $#xflog\n"; }
   @xflog=grep(!/\b$xname\b/i,@xflog);
  
 }

#STEP 2: foreach subsample compute the histogram of the coordinates and find peaks

my $nidx=-1;

#The following loop will:
# - Loop on the @onames array
# - Scan the @flog array for all the rows containing the current name
# - Build the PDL $coords containing the coords for these rows
# - Build clusters
# Output: @newtab: array of new names , coordinates  and indexes


@newtab=();
my @mra=();
my @mdec=();

foreach my $name (@onames)  {
$nidx++;

if ($DEBUG){  print "\n>> $name << \n"; }
  my @xflog=grep(/\b$name\b/i,@flog);
     my ($ra, $dec);
     my $nrows=$#xflog+1;
     my $coords=zeroes(float,$nrows,3);

      my $r=0;
     foreach my $row (@xflog) {
        my @fields=split(/$main::DBDEL/,$row);
        my $c=0;
        foreach my $field (@fields) {
           if ($header[$c] =~ 'Nr') {set $coords,$r, 0, $field}
           elsif ($header[$c] =~ 'RA') {my @str=split(/:/,$field,3);
                                        my $val=$str[0]+$str[1]/60.+$str[2]/3600.;
                                        set $coords,$r, 1, $val;
                                         }
           elsif ($header[$c] =~ 'DEC') {my @str=split(/:/,$field,3);
					my $sig=1;
					if ($str[0]=~/-/) {$sig=-1};
					my $val=$sig*(abs($str[0])+$str[1]/60.+$str[2]/3600.);
                                          set $coords,$r, 2, $val;
                                            last}
           $c++;
          } #end loop on fields
         $r++;
        } #end loop on rows of subsample

# print "   Mean coordinates:  \n";

           my $c1= $coords->slice(':,(1)') ;
           my $c2= $coords->slice(':,(2)') ;


	   my ($amin, $amax, $dmin, $dmax);
	   
	   $amin=min($c1); $amax=max($c1);
	   $dmin=min($c2); $dmax=max($c2);

	   
           my $ast=$thresh->at(0); 
           $amin-=$ast;
           $amax+=$ast;
          my $abins= int(($amax-$amin)/$ast+.5) ;

               my $dst=$thresh->at(1);
            $dmin-=$dst;
            $dmax+=$dst;
              my $dbins=int(($dmax-$dmin)/$dst+.5);

            #   print "$c1 $c2\n";
         if ($DEBUG){       print "$amin $amax $ast $abins, $dmin $dmax $dst $dbins\n";  }

#Compute the histogram of the coordinates and find clusters
         my $h = histogram2d($c1, $c2, $ast, $amin ,$abins , $dst, $dmin, $dbins);
        my ($nch, $nrh)=dims($h);
        my $nz=which($h->clump(2)>0);
        my $nzy=rint($nz/$nch);
        my $nzx=($nz-$nzy*$nch);

       # $clust=cat $amin-$ast+$nzx*$ast , $dmin-$dst+$nzy*$dst;
       # Either we should build an array of hashes with new names & coordinates
       # or all info should be written to a line.......
       # first better because then it is easy to reloop and look for misidentifications
       # structure: name, un/changed, alfa, delta
       # ($new->)
        my $mod='-'; #Default: all ok....
        if (nelem($nzx) > 1) {$mod='W'}  #more than one peak, something is suspicious......

       for my $clust (0..nelem($nzx)-1) {
          my $nname=$name;
             if ($clust>0) {$nname=$nname."#$clust"}
            my $clra= $amin+$ast*$nzx->at($clust);
            my $cldec=  $dmin+$dst*$nzy->at($clust);

             my $sc=which(($c1>= ($clra) & $c1 < ($clra+$ast)) & ($c2>= ($cldec) & $c2 < ($cldec+$dst))  );
          my $mra=median(index($c1,$sc));
          my $mdec=median(index($c2,$sc));
       if ($DEBUG) { print "$clra $ast $cldec $dst $sc  --> $mra $mdec\n";}
       push @newtab, [$nname, $nname, $mod, $mra, $mdec, index($coords->slice(':,(0)'),$sc)];
       push (@mra,$mra);
       push (@mdec,$mdec);
        }

  } #end loop on names

#convert to piddles and free memory
my $mra=float(@mra);
@mra=();
my $mdec=float(@mdec);
@mdec=();


#And now, we loop again and find different clusters with same coordinates..................
foreach my $i (0..$#newtab)  {
 # next if ($newtab[$i][2] ne 'W') ;  #uncomment to check only "suspicious" groups

   my $min_dra=9999;
   my $min_ddec=9999;
   my $guess=-1;
       foreach my $j (0..$#newtab)  {
          next if ($j==$i || $newtab[$j][2] ne '-'); #skip if the entry is the same or it is wrong 
           my $dra = abs($newtab[$i][3]- $newtab[$j][3] );
           my $ddec = abs( $newtab[$i][4]- $newtab[$j][4] );
          if ($dra <= $thresh->at(0) && $ddec <= $thresh->at(1))
                 {  if ($dra < $min_dra && $ddec < $min_ddec)  {
                          $min_dra=$dra; $min_ddec=$ddec; $guess=$j;$newtab[$i][2]='M';
                       }
    if ($DEBUG) {print "found !! $newtab[$i][0] & $newtab[$j][0]\n"}
        }
    }
          if ($guess>=0) {
              $newtab[$i][1]=$newtab[$guess][0];
              if ($DEBUG) {print ">>>>>> Most likely: $newtab[$guess][0]\n"}
            }
}

#The last effort: sort the @newtab array by alfa and delta

#First sort by ra
 my  $ira=qsorti($mra);

# now sort by dec

 my $rdim=nelem($mra);
 my $f=zeroes (ushort,$rdim);

$mra=$mra->index($ira);
$mdec=$mdec->index($ira);


my $i;

for ($i=0;$i<$rdim;$i++) {

#skip if flagged
  if ($f->at($i)==1) {next}

#find all obs with same ra and flag them
   my $sra=$mra->at($i);
   my $ridx=which ($mra==$sra);
   (my $tmp=$f->index($ridx)) .= 1;

#extract decs
   my $xdec=$mdec->index($ridx);
   my $ddim=nelem($xdec);

#skip if it is only one element
   if ($ddim<=1) {next}

#sort
   my $didx=qsorti($xdec);

#change the order of the sorting index
   ($tmp = $ira->index($ridx)) .= $tmp->index($didx);

}



#interactively change @newtab

 my $atab = $main::top->Toplevel;
$atab->title('Clusters of names');

# Use a scrollable table (useful for task with many parameters)

 my $at  = $atab->Table(-scrollbars => 'e',-fixedrows => 1, -rows => 20);
$at->pack(-expand=> 1, -fill => 'both');

my $c=0;

$at->put(0,$c++,$at->Label(-text=>"Name",-anchor=>'w'),-relief => 'raised');
$at->put(0,$c++,$at->Label(-text=>"RA",-anchor=>'w'),-relief => 'raised');
$at->put(0,$c++,$at->Label(-text=>"DEC",-anchor=>'w'),-relief => 'raised');
$at->put(0,$c++,$at->Label(-text=>"Flag",-anchor=>'w'),-relief => 'raised');

$at->configure(-columns => $c) if ($c > $at->cget('-columns'));

my $buttons = $atab->Frame();
$buttons->pack(-side => 'bottom', -pady => '2', -anchor => 'center');
#my $buttons_clear = $buttons->Button(-text => 'Clear');
#my $buttons_load = $buttons->Button(-text => 'Load');
my $buttons_apply = $buttons->Button(-text => 'Apply');
$buttons_apply->configure(-command => \&apply_changes);
#my $buttons_savedef = $buttons->Button(-text => 'Save def.');
my $buttons_exit = $buttons->Button(-text => 'Close');
$buttons_exit->configure(-command =>  (sub {$atab->destroy}));
#$buttons_skip = $buttons->Button(-text => 'Next');

#If not interactive mode, windows are only used to show the parameters (read only)
#if ($main::redmode eq "interactive") {
#    $buttons_clear->pack(-side => 'left', -padx => '2');
#    $buttons_load->pack(-side => 'left', -padx => '2');
    $buttons_apply->pack(-side => 'left', -padx => '2');
#    $buttons_savedef->pack(-side => 'left', -padx => '2');
    $buttons_exit->pack(-side => 'left', -padx => '2');
    #$buttons_skip->pack(-side => 'left', -padx => '2');
#  }


foreach my $i (0..$#newtab)  {
  my $j=$ira->at($i);

       my $xpos=0;

       my $e;

            ${$e}=$at->BrowseEntry(-width =>'30', -variable => \$newtab[$j][1]);
         $frames[$i+1]=${$e};$at->put($i+1,$xpos++,${$e});
         if ($newtab[$j][1] ne $newtab[$j][0]) {$frames[$i+1]->insert("end",$newtab[$j][1])}
         $frames[$i+1]->insert("end", $newtab[$j][0]);

        foreach my $fc (0..1) {
           my $field=$newtab[$j][$fc+3];
            my $sfield=sexstr($field);
            $newtab[$j][$fc+3]= $sfield;
           $at->put($i+1,$xpos++,$at->Label(-text => "$sfield", -anchor => 'e'));
        }

         $at->put($i+1,$xpos++,$at->Label(-text => "$newtab[$j][2]", -anchor => 'e'));

}

$frames[1]->focus;

$atab->waitWindow;

}

sub apply_changes
{


#replace names and coords in the dblog array (should be the last step....)
foreach my $i (0..$#newtab)  {
  #foreach my $j (0..$#{$newtab[$i]}-1) {
     #        if ($DEBUG) {print "$i $j  $newtab[$i][$j] ;"}
   #}
   my $p=$newtab[$i][$#{$newtab[$i]}] ;

#update the database
   foreach my $j (0..nelem($p)-1) {
     my $idx=$p->at($j);
             if ($DEBUG) {print ">>>>>>> Reading $idx ($newtab[$i][1])/ ";}
     my $row= $log[$idx];
     my @fields=split(/$main::DBDEL/,$row);
        my $c=0;
        my $nrow='';
	my $oname="";

        foreach my $field (@fields) {
           if ($header[$c] =~ "object") {$oname=$field;$oname =~ tr/ //d;
	   $nrow=$nrow."$newtab[$i][1]".$main::DBDEL}
            elsif ($header[$c] =~ "RA") {$nrow=$nrow."$field".$main::DBDEL}   #we display the TRUE coordinates 
            elsif ($header[$c] =~ "DEC") {$nrow=$nrow."$field".$main::DBDEL}
	    elsif ($header[$c] =~ "Warnings") {
	      # chomp ($field);    #remove the \n
	       
	       #my $comm="\t Wrong name";
	       #if ($newtab[$i][2] eq 'W' || $newtab[$i][2] eq 'M') {$comm=$comm.'??'}
	       #elsif  ($newtab[$i][1] ne $oname) {
	       #         $comm=$comm.": name changed (was $oname)"
	       #}
	       #else {$comm=''};
	       
	       my $comm='';
	       if ($newtab[$i][1] ne $oname) {$comm="# Name changed (was $oname)"}
	       
	       $nrow=$nrow.$field.$comm.$main::DBDEL; 	       
	     }
           else {$nrow=$nrow.$field.$main::DBDEL}
           $c++;
          } #end loop on fields
	  if ($DEBUG) {print "\n###### idx -  $log[$idx] -> $nrow ##### \n"};

       $log[$idx]=$nrow ; 
}

if ($DEBUG) {print "\n";}

}

#ok, dump the new @log to the file

open (NEWDB, ">$odbfile");

foreach my $row (@log) {chomp($row); print NEWDB "$row\n"; }

close (NEWDB);

$main::top->Dialog(
       -title  => 'CheckDB',
       -text   => "The file $odbfile has been made",
       -bitmap => 'warning',
   )->Show;

}

sub sexstr {
   my $value=shift;
   my $f1s='';
   my $f1=int($value);
   $f1s= '-' if ($f1 <0);
   my $f3=($value-$f1)*60;
   if ($f3 <0) {$f1s='-';$f3*=-1}
   my $f2=int($f3);
     $f3=int( ($f3-$f2)*60.*100+.5)/100.;
   if ($f3>=60) {$f2++; $f3-=60}
   $f3=sprintf("%.2f",$f3);


   $f1='0'.abs($f1) if (abs($f1<10));
   $f2='0'.$f2 if ($f2<10);
   $f3='0'.$f3 if ($f3<10);

   $value=$f1s.$f1.':'.$f2.':'.$f3;

   return($value);
}

1;
