use Cwd;

use vars qw/@flist $treetop/;
use strict;

use Tk::DirTree;

sub MakeLinks {

  my $mklWdw= $main::top->Toplevel;
  $mklWdw->title('Make Links');
 
  my $fr = $mklWdw->Frame();
  $fr->pack();;

  my $lins= $fr->Label(-text => "Instrument: ");
  my $e_ins=$fr->BrowseEntry(-choices=>$main::IMPL_INSTR,
       -variable => \$main::INSTRUMENT, -state=>'readonly');
  
  my $ldbf= $fr->Label(-text => "DataBase file");
  my $e_dbf=$fr->Entry(-textvariable => \$main::SDFILE, -width=>'30');
 
 my $buttons_dbf = $fr->Button(-text => 'Browse...',
      -command => (sub { &getFile('load',\$main::SDFILE, '.db','Select the INPUT file')} )
   );

    
     my $lid= $fr->Label(-text => "Input Directory");
    my $e_id= $fr->Entry(-textvariable => \$main::DBINDIR, -width=>'30');
    my $buttons_indir = $fr->Button(-text => 'Set Input Directory',
       -command => (sub {&getTree(\$main::DBINDIR, 
            'Please select the directory containing the INPUT data') } )
      );    
    
   my $lod= $fr->Label(-text => "Output Directory");
    my $e_od= $fr->Entry(-textvariable => \$main::DATADIR, -width=>'30');
   my $buttons_outdir = $fr->Button(-text => 'Set Output Directory',
   -command => (sub {&getTree(\$main::DATADIR, 
      'Please select the directory where links should be created') } )
   );

  
   
my ($mkDB, $mkLinks);

$mkDB=1; $mkLinks=0;


my $bmkl= $fr->Checkbutton(
     -text     => "Make the LINKS",
     -variable => \$mkLinks,
     -relief   => 'flat',
 );



my $bmkdb= $fr->Checkbutton(
     -text     => "Build the DATABASE",
     -variable => \$mkDB,
     -relief   => 'flat',
 );


      my $buttons_OK = $fr->Button(-text => 'GO',
         -command => ( sub { 
	  &mktree($main::DBINDIR,$main::DATADIR, $main::SDFILE, $main::INSTRUMENT, $mkDB,$mkLinks  ) ;
         }));


     my $buttons_close = $fr->Button(-text => 'Close',
     -command => (sub {$mklWdw->destroy}));
    
    my $buttons_save = $fr->Button(-text => 'Save to prefs.',
     -command => (sub {&SavePrefs()}));


$lins->grid($e_ins);
$ldbf->grid($e_dbf,$buttons_dbf);
$lid->grid($e_id,$buttons_indir);
$lod->grid($e_od,$buttons_outdir);

$bmkl->grid($bmkdb);
$buttons_OK->grid($buttons_close, $buttons_save);

}

sub mktree {
my ($indir, $outdir, $database, $INSTRUMENT, $mkDB, $mkLinks)=@_;

return if (!$mkDB && !$mkLinks);

    if (!-d $indir) {
	$main::top->Dialog(
            -title  => 'ERROR !!!',
            -text   => "Can't find dir $indir",
            -bitmap => 'error',
        )->Show;
	return;
     }


#===> we want to be sure that the name of the file is the FILENAME stored in the header

my $hinfo;

my $tname=lc($INSTRUMENT);
if (! eval "require Instr::$tname") {
  print "instrument $tname not defined !\n";
  return -1;
}

eval "\$hinfo=$tname->new()";

$hinfo->init_head();    #Set all parameters (header)
$hinfo->init_props();    #Set all parameters (properties)


if ($mkLinks) {
 system ("mkdir $outdir") if (!-d $outdir);
    
 if (!-d $outdir) {
	$main::top->Dialog(
            -title  => 'ERROR !!!',
            -text   => "Can't write to dir $outdir",
            -bitmap => 'error',
        )->Show;
	return
     }
}


if (!open(I_FOUND,"find $indir -name '*.fits' -print |")) {
   $main::top->Dialog(-title  => 'ERROR !!!',
            -text   => "the find command failed ?!?",
            -bitmap => 'error',
        )->Show;
	return
 }


if ($mkDB) {
   my $now=&main::mkdate;
   open (DBFILE, "> $database");
   print DBFILE "<DATE> $now </DATE>\n";

   print DBFILE "<HEADER>\n";
      my $leg=join $main::DBDEL, 'Nr', $hinfo->dbKeys, 'Extensions', 'Warnings';
      print DBFILE "$leg\n";
   print DBFILE "</HEADER>\n";
}

my $seqnr=1;


my $NAME;
NAME: while ($NAME = <I_FOUND>) 
{
  chomp($NAME);
  if( -f $NAME) {
    $hinfo->Translate($NAME);
     my $minccd=$hinfo->getProps('MINCCD');
     
  if ($mkLinks) {
#    my $PNAME=$hinfo->FILENAME;   #the original name in the header

#    if ($PNAME eq 'UNDEFINED') {
#        my @s=split(/\//,$NAME); 
#	$PNAME=$s[-1]
#    }

# ---------- The name of the output link
#    my $lkname=$outdir.'/'.$PNAME;   # the link

     my $lkname=$outdir.'/';
     my @allsubdir;

#In the case of SIMPLE files, we must add a directory


     if  ($hinfo->getProps('FITS_TYPE') eq 'SIMPLE') {
          my $PNAME=$hinfo->FILENAME;
          if ($PNAME ne 'UNDEFINED') {
	    $lkname.=$PNAME;
	  } else {
	    my @s= split(/\//,$NAME);
	    @s=split(/\./, $s[-1]);
	    $lkname.= '/'.$s[0];
	  }
	    my @s= split(/\//,$NAME);
	    $lkname.='/'.$s[-1];
	    @allsubdir=split(/\//,$lkname);  
            } else {
	      @allsubdir=split(/\//,$lkname);  
	      my @s= split(/\//,$NAME);
	      $lkname.=$s[-1];
	    }


#warn "should link $NAME to $lkname";

   if (! -e $lkname) {

     #my $END_NAME=pop @allsubdir;
     my $topdir='';
 
SDIR: foreach my $sdir (@allsubdir) {
       $topdir.=$sdir.'/';
         if (! -d $topdir) {system ("mkdir $topdir\n"); print "$topdir created\n"}
     }

    if (! symlink ($NAME, $lkname)) {warn "could not link $lkname to $NAME"} 
   }
  
   } #end if mkLinks
   
   if ($mkDB) {
       my @warnings=();
       #if ($END_NAME !~ /\A[0-9]/ ) {my $comm='Average ?'; push(@warnings,$comm) }
       if ($hinfo->getProps('FITS_TYPE') eq 'SIMPLE') {                   
               next NAME if $hinfo->IMAGEID >  $minccd;
        } 
# ========= MAKE SOME (SIMPLE) CONSISTENCY CHECKS AND ISSUE WARNINGS ==========	
        my $obstype=$hinfo->OBSTYPE;
	my $exptime=$hinfo->EXPTIME;
	if ($obstype eq "DARK" or $obstype eq "BIAS") {
           #if BIAS or DARK, check for exposure time
           # (if Dark with 0s, it is a bias; conversely, if Bias with not 0s it is a Dark)
            
	    if ($exptime == 0) {
                  if ($obstype ne "BIAS") {my $comm="# $obstype changed to BIAS";push(@warnings,$comm)}
	                                     $obstype = "BIAS"}
                 else {if ($exptime != 0) {
                 if ($obstype ne "DARK") {my $comm="# $obstype changed to DARK";push(@warnings,$comm)}
                                             $obstype = "DARK"}
	             }
	       
              $hinfo->OBJECT($obstype);
              $hinfo->FILTER("");
           } 
	if ($exptime ==0 &&$obstype ne "BIAS" ) {
	    my $comm="# $obstype changed to BIAS";push(@warnings,$comm);
	     $obstype = "BIAS";
	     $hinfo->OBJECT($obstype);
	     $hinfo->FILTER("");
	}
# ====== END	   
	my $nfound=$hinfo->HowMany($NAME);             
        my $row=$hinfo->mkSummary($main::DBDEL);
#warn "--- $NAME $row" if $nfound==0;
	if ($row ne '') { 
#warn "inserting data for $NAME";  
	    $row=join "$main::DBDEL", $seqnr, $row, $nfound, @warnings;
            print DBFILE "$row\n";
	    $seqnr++;
         }     
             
   } #end if mkDB
  } #end if file exists
} #end while
close (I_FOUND);


close DBFILE if ($mkDB);

return;


} 

sub get_select {

 
 my ($path, @input)=@_;

    while ($_ = shift @input) {
      my $file=$path;
      if ($_ ne '.') {$file=$file."/$_"}      
        #print ".... $file\n";
       if (-d $file) {
	     push(@flist,$file);
#It is a directory, looking inside.....
           opendir H, $file;
	    my @newlev= grep ! /^\.\.?$/,sort(readdir H);
	   closedir H;
           get_select($file,@newlev);
         }
    }    
} #end get_select



sub getTree {

my   $seldir=shift;
my $message=shift;

my $current_dir=$$seldir;
   if (not $current_dir) {$current_dir=cwd;}
my $done=0;

if (not -d $current_dir) {$current_dir = cwd};

if (not Exists $treetop) {
              
   $treetop= $main::top->Toplevel;
   $treetop->title('Select a directory');

#   my $tree = $treetop->Scrolled( qw/DirTree
#                               -width 35 -height 30
#         -selectmode browse -exportselection 1
#               -scrollbars osoe/ );

  my $tree = $treetop->Scrolled( qw/DirTree
                                -width 35 -height 30
                        -exportselection 1
              -scrollbars osoe/ );

    my $lab = $treetop->Label( -text => $message );
    my $ent = $treetop->Entry( -textvariable => \$current_dir );
    my $info= $treetop->Label( -text => "You can also type here the path" );
 
    my $ok = $treetop->Button( qw/-text OK -underline 0 -width 6/ );

    my $cancel = $treetop->Button( qw/-text Cancel -underline 0 -width 6/ );

    $tree->configure(   -browsecmd => sub { $current_dir = shift } );
# this means that double clicking has the same effect as the OK button
    $tree->configure(   -command   => sub {$$seldir=$current_dir; $done=1;$treetop->destroy} );                      


    $ok->configure(     -command   => sub { $$seldir=$current_dir; $done=1;$treetop->destroy});

    $cancel->configure( -command   => sub { $treetop->destroy } );
 
    $lab->pack(    qw/-anchor w/ );

    $tree->pack(   qw/-expand yes -fill both -padx 10 -pady 10 -side top/ ); 

    $info->pack(    qw/-anchor w/ );  
    $ent->pack(    qw/-fill x/ );

    $ok->pack(     qw/-side left  -padx 10 -pady 10/ );
    $cancel->pack( qw/-side right -padx 10 -pady 10/ );
						
 }

$treetop->waitWindow;

return $done;

} #end getTree


1;
